# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Streamlit widgets for simulation web UI."""

import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from concurrent.futures import Future
from typing import Optional

import altair as alt
import humanize
import pandas as pd
import streamlit as st
from core.sim_time import TimeUnit, ensure_time
from core.utils import get_simulation_stats
from numpy.typing import NDArray
from streamlit.delta_generator import DeltaGenerator

from streaming.base.util import bytes_to_int


def get_line_chart(data: pd.DataFrame,
                   throughput_window: int,
                   throughput: bool = True) -> alt.Chart:
    """Get interactive line charts for throughput or cumulative downloads.

    Args:
        data (pd.DataFrame): Dataframe containing throughput or cumulative downloads data.
        throughput_window (int): Window size for throughput rolling average.
        throughput (bool, optional): Whether to display throughput or cumulative downloads. If
            ``True``, returns chart for throughput. Otherwise, returns chart for downloads.
            Defaults to ``True``.

    Returns:
        alt.Chart: Interactive line chart for throughput or cumulative downloads.
    """
    hover = alt.selection_point(
        fields=['step'],
        nearest=True,
        on='mouseover',
        empty=False,
    )

    lines = (alt.Chart(data,
                       title='Throughput (' + str(throughput_window) +
                       '-step rolling average)').mark_line().encode(
                           x='step',
                           y='throughput (batches/s)',
                       )) if throughput else (alt.Chart(
                           data, title='Cumulative Network Usage, all nodes').mark_line().encode(
                               x='step', y='cumulative network usage (bytes)'))

    # Draw points on the line, and highlight based on selection
    points = lines.transform_filter(hover).mark_circle(size=65)

    # Draw a rule at the location of the selection
    tooltips = (
        alt.Chart(data).mark_rule().encode(
            x='step',
            y='throughput (batches/s)' if throughput else 'cumulative network usage (bytes)',
            opacity=alt.condition(hover, alt.value(0.3), alt.value(0)),  # pyright: ignore
            tooltip=[
                alt.Tooltip('step', title='Step'),
                alt.Tooltip(
                    'throughput (batches/s)' if throughput else 'cumulative network usage (bytes)',
                    title='Throughput' if throughput else 'Network Usage'),
            ],
        ).add_params(hover))
    return (lines + points + tooltips).interactive()  # pyright: ignore


def stream_entry(component: DeltaGenerator,
                 streams: dict,
                 key: int,
                 add_stream: bool = True,
                 defaults: Optional[dict] = None):
    """Define stream input area widget.

    Args:
        component (DeltaGenerator): Streamlit component to write on.
        streams (dict): Dictionary to store different stream entries.
        key (int): Key for stream entry.
        add_stream (bool, optional): Whether to show the option to add another stream.
            Defaults to ``True``.
        defaults (dict, optional): Dictionary of default values for stream entries. Defaults to
            ``None``.
    """
    stream_entries = {}
    component.write(f'*Stream {key+1}*')
    on = component.toggle('use `index.json`', key=str(key) + 'toggle') if add_stream else None
    if on or not add_stream:
        path = component.text_input(
            'path to `index.json`',
            value='/absolute/path/to/index.json' if defaults is None else defaults['path'],
            help='path to the `index.json` file for this stream. \
                                the `index.json` file contains information about the shards in \
                                your dataset.',
            key=str(key) + 'path',
            disabled=(not add_stream))
        if add_stream:
            path_type = component.selectbox('path type', ['local', 'remote'],
                                            key=str(key) + 'path_type')
            stream_entries['path_type'] = path_type
        if path[-1] != '/':
            path += '/'
        stream_entries['path'] = path
    else:
        shards = component.number_input('number of shards',
                                        step=1,
                                        value=2000,
                                        help='number of \
                                        total shards across your whole dataset.',
                                        key=str(key) + 'shards')
        samples_per_shard = component.number_input('samples per shard',
                                                   step=1,
                                                   value=4000,
                                                   help='average number of samples contained \
                                                in each shard.',
                                                   key=str(key) + 'samples')
        avg_raw_shard_size = component.text_input('avg raw shard size (bytes)',
                                                  value='60MB',
                                                  help='average raw size, in bytes, \
                                            of a single shard.',
                                                  key=str(key) + 'rawsize')
        avg_raw_shard_size = bytes_to_int(avg_raw_shard_size)
        avg_zip_shard_size = component.text_input('avg compressed shard size (bytes)',
                                                  value='None',
                                                  help='average compressed size, \
                                                    in bytes, of a single shard.',
                                                  key=str(key) + 'zipsize')
        avg_zip_shard_size = None if avg_zip_shard_size == 'None' \
            else bytes_to_int(avg_zip_shard_size)
        stream_entries['shards'] = shards
        stream_entries['samples_per_shard'] = samples_per_shard
        stream_entries['avg_raw_shard_size'] = avg_raw_shard_size
        stream_entries['avg_zip_shard_size'] = avg_zip_shard_size
    proportion = component.text_input(
        'proportion',
        value='None' if defaults is None else str(defaults['proportion']),
        help='proportion of the full training dataset that this stream \
                                    represents.',
        key=str(key) + 'proportion',
        disabled=(not add_stream))
    proportion = None if proportion == 'None' or proportion == '' else float(proportion)
    repeat = component.text_input('repeat',
                                  value='None' if defaults is None else str(defaults['repeat']),
                                  help='number of times to repeat the samples in this \
                                stream.',
                                  key=str(key) + 'repeat',
                                  disabled=(not add_stream))
    repeat = None if repeat == 'None' or repeat == '' else float(repeat)
    choose = component.text_input('choose',
                                  value='None' if defaults is None else str(defaults['choose']),
                                  help='number of samples to choose from this \
                                stream.',
                                  key=str(key) + 'choose',
                                  disabled=(not add_stream))
    choose = None if choose == 'None' or choose == '' else int(choose)
    stream_entries['proportion'] = proportion
    stream_entries['repeat'] = repeat
    stream_entries['choose'] = choose

    streams[key] = stream_entries
    if add_stream and component.checkbox(label='add stream', key=str(key) + 'checkbox'):
        stream_entry(component, streams, key + 1)


def param_inputs(component: DeltaGenerator, input_params: dict, defaults: dict = {}):
    """Define parameter input area widget.

    Args:
        component (DeltaGenerator): Streamlit component to write to.
        input_params (dict): Dictionary to store input parameters.
        defaults (dict): Dictionary of default values for input params. Defaults to empty dict, {}.
    """
    # split the input column component into left, middle, and right sub columns.
    col_l, col_m, col_r = component.columns(3)

    # dataset
    streams = {}
    col_l.write('**Dataset Parameters**')
    if 'streams' in defaults:
        key = 0
        for stream in defaults['streams'].values():
            # Case is only possible when reading in streams from yaml file. Stream will have path.
            stream_entry(col_l, streams, key, add_stream=False, defaults=stream)
            key += 1
        streams = defaults['streams']
    else:
        stream_entry(col_l, streams, 0, add_stream=True)
    col_l.text('')
    input_params['streams'] = streams

    # training
    col_m.write('**Training Parameters**')
    if 'max_duration' in defaults:
        default_max_duration = defaults['max_duration']
        default_value = int(default_max_duration.value)
        default_unit_index = 0 if default_max_duration.unit == TimeUnit.BATCH else 1
        time_value = col_m.number_input('training duration',
                                        step=1,
                                        value=default_value,
                                        help='training duration value, in specified units.')
        time_units = col_m.selectbox('units', ['batches', 'epochs'],
                                     index=default_unit_index,
                                     help='units of training duration.')
    else:
        time_value = col_m.number_input('training duration',
                                        step=1,
                                        value=1000,
                                        help='training duration value, in specified units.')
        time_units = col_m.selectbox('units', ['batches', 'epochs'],
                                     help='units of training duration.')
    # Get Time object from inputs
    time_string = str(time_value)
    time_string += 'ba' if time_units == 'batches' else 'ep'
    max_duration = ensure_time(time_string, TimeUnit.EPOCH)
    epoch_size = col_m.text_input('epoch size (samples)',
                                  value='',
                                  help='epoch size for this run, in samples.')
    epoch_size = None if epoch_size == '' or epoch_size == 'None' else int(epoch_size)
    device_batch_size = col_m.number_input(
        'device batch size',
        step=1,
        value=16 if 'device_batch_size' not in defaults else defaults['device_batch_size'],
        help='number of samples per device (GPU) per batch. \
                                            the global batch size is `device_batch_size * \
                                            devices_per_node * physical_nodes`')
    col_m.text('')
    input_params['max_duration'] = max_duration
    input_params['epoch_size'] = epoch_size
    input_params['device_batch_size'] = device_batch_size

    # hardware and network
    col_m.write('**Hardware and Network Parameters**')
    physical_nodes = col_m.number_input(
        'number of physical nodes',
        step=1,
        value=1 if 'physical_nodes' not in defaults else defaults['physical_nodes'],
        help='number of physical nodes for this run. \
                                        a node typically consists of 8 devices (GPUs).')
    devices = col_m.number_input('devices per node',
                                 step=1,
                                 value=8 if 'devices' not in defaults else defaults['devices'],
                                 help='number of devices (GPUs) per node for this run. \
                                    there are typically 8 devices per node.')
    time_per_sample = col_m.number_input(
        'process time per sample (s)',
        step=0.0005,
        value=0.0175 if 'time_per_sample' not in defaults else defaults['time_per_sample'],
        format='%.4f',
        help='time for one device to process one \
                                        sample from your dataset.')
    node_network_bandwidth = col_m.text_input(
        'network bandwidth per node (bytes/s)',
        value='500MB'
        if 'node_network_bandwidth' not in defaults else defaults['node_network_bandwidth'],
        help='network bandwidth available to \
                                            each node. in practice, network bandwidth is \
                                            variable and is affected by many factors, \
                                            including cluster demand.')
    col_m.text('')
    input_params['physical_nodes'] = physical_nodes
    input_params['devices'] = devices
    input_params['time_per_sample'] = time_per_sample
    input_params['node_network_bandwidth'] = node_network_bandwidth

    # streaming
    col_r.write('**Streaming Parameters**')
    workers = col_r.number_input('workers per device',
                                 step=1,
                                 value=8 if 'workers' not in defaults else defaults['workers'],
                                 help='number of dataloader workers per device (GPU).')
    canonical_nodes = col_r.text_input(
        'number of canonical nodes',
        value='None' if 'canonical_nodes' not in defaults else defaults['canonical_nodes'],
        help='number of canonical nodes to split your dataset \
                                            into. a canonical node is a bucket of shards that is \
                                            assigned to a particular physical node.')
    canonical_nodes = None if canonical_nodes == '' or canonical_nodes == 'None' \
        else int(canonical_nodes)
    predownload = col_r.text_input(
        'predownload per worker (samples)',
        value='None' if 'predownload' not in defaults else defaults['predownload'],
        help='number of samples ahead each worker should download. \
                                    predownload does not occur before the first batch; \
                                    rather, it occurs while training is ongoing.')
    predownload = None if predownload == '' or predownload == 'None' else int(predownload)
    shuffle = col_r.checkbox(label='shuffle',
                             value=True if 'shuffle' not in defaults else defaults['shuffle'],
                             help='whether or not to shuffle the samples for this run.')
    shuffle_algo='py1e' if len(defaults) == 0 or 'shuffle_algo' not in defaults \
        else defaults['shuffle_algo']
    shuffle_block_size=None if len(defaults) == 0 or 'shuffle_block_size' not in defaults \
        else defaults['shuffle_block_size']
    seed = 42 if len(defaults) == 0 or 'seed' not in defaults else defaults['seed']
    if shuffle:
        algos = ['py1e', 'py1br', 'py1b', 'py1s', 'py2s', 'naive']
        default_index = 0
        if 'shuffle_algo' in defaults:
            default_index = algos.index(defaults['shuffle_algo'])
        shuffle_algo = col_r.selectbox(
            'shuffling algorithm',
            algos,
            index=default_index,
            help='shuffling algorithm to use for this run. your shuffle \
                                        parameters may affect model training.')
        shuffle_block_size = col_r.text_input(
            'shuffle block size (samples)',
            value='None'
            if 'shuffle_block_size' not in defaults else defaults['shuffle_block_size'],
            help='shuffle block size for this run. used in the `py1b`, `py1br`, and `py1e` \
                shuffling algorithms, samples in blocks of `shuffle_block_size` are randomly \
                shuffled inside each bucket of shards (aka canonical node).')
        seed = col_r.number_input('shuffle seed',
                                  step=1,
                                  value=42 if 'seed' not in defaults else defaults['seed'],
                                  help='random seed for shuffling.')
        if shuffle_block_size == '' or shuffle_block_size == 'None':
            shuffle_block_size = None
    cache_limit = col_r.text_input(
        'cache limit (bytes)',
        value='None' if 'cache_limit' not in defaults else defaults['cache_limit'],
        help='cache limit per node for this run. \
                                    setting cache limit too low will impact throughput.')
    cache_limit = None if cache_limit == '' or cache_limit == 'None' else bytes_to_int(cache_limit)
    sampling_methods = ['balanced', 'fixed']
    sampling_method = col_r.selectbox('sampling method',
                                      sampling_methods,
                                      index=0 if 'sampling_method' not in defaults else
                                      sampling_methods.index(defaults['sampling_method']),
                                      help="sampling method for this run. controls how samples are\
                                        chosen each epoch. can be either 'balanced' or 'fixed'.")
    sampling_granularity = col_r.number_input(
        'sampling granularity',
        step=1,
        value=1 if 'sampling_granularity' not in defaults else defaults['sampling_granularity'],
        help='sampling granularity for this run. controls how\
                                        samples are balanced across shards. higher values will\
                                        cause more samples to be drawn from each shard at a time.')
    batching_methods = ['random', 'per_stream', 'stratified']
    batching_method = col_r.selectbox('batching method',
                                      batching_methods,
                                      index=0 if 'batching_method' not in defaults else
                                      batching_methods.index(defaults['batching_method']),
                                      help='batching method for this run. controls how batches\
                                        are constructed.')
    col_r.text('')
    input_params['workers'] = workers
    input_params['canonical_nodes'] = canonical_nodes
    input_params['predownload'] = predownload
    input_params['cache_limit'] = cache_limit
    input_params['shuffle'] = shuffle
    input_params['shuffle_algo'] = shuffle_algo
    input_params['shuffle_block_size'] = shuffle_block_size
    input_params['seed'] = seed
    input_params['sampling_method'] = sampling_method
    input_params['sampling_granularity'] = sampling_granularity
    input_params['batching_method'] = batching_method


def display_simulation_stats(component: DeltaGenerator, total_batches: int, step_times: NDArray,
                             time_per_sample: float, device_batch_size: int, startup_time: float,
                             min_cache_limit: int, cache_limit: Optional[int]):
    """Display simulation statistics and warnings.

    Args:
        component (DeltaGenerator): Streamlit component to write on.
        total_batches (int): Total number of batches in the simulation.
        step_times (NDArray): Array of step times from the simulation.
        time_per_sample (float): Time taken for one device to process one sample, in seconds.
        device_batch_size (int): Device batch size.
        startup_time (float): startup time from the simulation.
        min_cache_limit (int): Minimum cache limit needed to prevent shard redownloads.
        cache_limit (Optional[int]): Cache limit provided.
    """
    all_throughput_drops, warmup_time, warmup_step, post_warmup_throughput_drops = \
        get_simulation_stats(step_times, time_per_sample, device_batch_size)
    with component.container():
        st.write(f'Minimum cache limit needed: **{humanize.naturalsize(min_cache_limit)}**')
        if cache_limit is not None and cache_limit < min_cache_limit:
            # Cache limit is too low, and will cause shard redownloads / throughput drops.
            st.warning('The provided cache limit is lower than the minimum cache limit needed to \
                     prevent shard re-downloads. This can cause throughput issues.',
                       icon='⚠️')
        if warmup_step == total_batches:
            # display error if the warmup phase is the whole run,
            # meaning that we never hit peak throughput.
            st.error('This configuration is severely bottlenecked by downloading. \
                     The run will not be performant.',
                     icon='🚨')
        elif post_warmup_throughput_drops:
            # display warning if post-warmup throughput drops are more than 10% of the run.
            st.warning('This configuration experiences some downloading-related slowdowns \
                       even after warmup.',
                       icon='⚠️')
        st.write('**{0} steps**, or **{1:.1f}%** of all steps, waited for \
                 shard downloads.'.format(all_throughput_drops,
                                          100 * all_throughput_drops / (total_batches)))
        if warmup_step != total_batches:
            # only display post-warmup throughput drop info if we actually ended the warmup period
            # (i.e. we hit peak throughput at some point)
            st.write('There were **{} steps** that waited for shard downloads after the warmup \
                     period.'.format(post_warmup_throughput_drops))
        st.write('Estimated time to first batch: **{0:.2f} s**'.format(startup_time))
        st.write('Estimated warmup time: **{0:.2f} s**'.format(warmup_time))


def get_shuffle_quality_chart(data: pd.DataFrame) -> alt.Chart:
    """Get interactive bar chart for shuffle quality.

    Args:
        data (pd.DataFrame): Dataframe containing shuffle quality data.

    Returns:
        alt.Chart: Interactive bar chart for shuffle quality.
    """
    bars = (alt.Chart(data,
                      title='Relative Shuffle Quality (1 is best)').mark_bar().encode(
                          x='shuffling algorithm',
                          y='relative quality',
                          tooltip='relative quality').properties(width=550,))
    return bars.interactive()


def display_shuffle_quality_graph(futures: list[Future], component: DeltaGenerator):
    """Display shuffle quality graph.

    Args:
        futures (list[Future]): List of futures for shuffle quality results.
        component (DeltaGenerator): Streamlit component to write on.
    """
    # Retrieve shuffle quality result since it is available
    shuffle_algos_qualities = list(zip(*[f.result() for f in futures]))
    shuffle_algos = list(shuffle_algos_qualities[0])
    # divide all shuffle qualities by naive shuffle quality to get a relative measure
    naive_idx = shuffle_algos.index('naive')
    naive_shuffle_quality = shuffle_algos_qualities[1][naive_idx]
    shuffle_algos.remove('naive')
    shuffle_qualities = [
        shuffle_algos_qualities[1][i] / naive_shuffle_quality
        for i in range(len(shuffle_algos_qualities[1]))
        if i != naive_idx
    ]
    shuffle_quality_df = pd.DataFrame({
        'shuffling algorithm': shuffle_algos,
        'relative quality': shuffle_qualities
    })
    component.altair_chart(get_shuffle_quality_chart(shuffle_quality_df), use_container_width=True)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Barrier that lives in shared memory.

Implemented with shared array and a filelock.
"""

from time import sleep

import numpy as np
from filelock import FileLock

from streaming.base.constant import TICK
from streaming.base.shared.array import SharedArray

# Time out to wait before raising exception
TIMEOUT = 60


class SharedBarrier:
    """A barrier that works inter-process using a filelock and shared memory.

    We set the number of processes (and thereby initialize num_exit) on the first time this object
    is called. This is because the object is created in a per-rank process, and called by worker
    processes.

    Args:
        filelock_path (str): Path to lock file on local filesystem.
        shm_name (str): Shared memory object name in /dev/shm.
    """

    def __init__(self, filelock_path: str, shm_name: str) -> None:
        # Create lock.
        self.filelock_path = filelock_path
        self.lock = FileLock(self.filelock_path)

        # Create three int32 fields in shared memory: num_enter, num_exit, flag.
        self._arr = SharedArray(3, np.int32, shm_name)
        self.num_enter = 0
        self.num_exit = -1
        self.flag = True

    @property
    def num_enter(self) -> int:
        """Get property num_enter.

        Returns:
            int: Number of processes that have entered the barrier.
        """
        return self._arr[0]

    @num_enter.setter
    def num_enter(self, num_enter: int) -> None:
        """Set property num_enter.

        Args:
            num_enter (int): Number of processes that have entered the barrier.
        """
        self._arr[0] = num_enter

    @property
    def num_exit(self) -> int:
        """Get property num_exit.

        Returns:
            int: Number of processes that have exited the barrier.
        """
        return self._arr[1]

    @num_exit.setter
    def num_exit(self, num_exit: int) -> None:
        """Set property num_exit.

        Args:
            num_exit (int): Number of processes that have exited the barrier.
        """
        self._arr[1] = num_exit

    @property
    def flag(self) -> bool:
        """Get property flag.

        Returns:
            bool: The flag value.
        """
        return bool(self._arr[2])

    @flag.setter
    def flag(self, flag: bool) -> None:
        """Set property flag.

        Args:
            flag (bool): The flag value.
        """
        self._arr[2] = bool(flag)

    def __call__(self, num_procs: int) -> None:
        """A set number of processes enter, wait, and exit the barrier.

        Args:
            num_procs (int): How many processes are sharing this barrier.
        """
        # Initialize num_exit to the number of processes.
        with self.lock:
            if self.num_exit == -1:
                self.num_exit = num_procs

        # If we are the first to arrive, wait for everyone to exit, then set flag to "don't go".
        self.lock.acquire()
        if not self.num_enter:
            self.lock.release()
            while self.num_exit != num_procs:
                sleep(TICK)
            self.lock.acquire()
            self.flag = False

        # Note that we entered.
        self.num_enter += 1

        # If we are the last to arrive, reset `enter` and `exit`, and set flag to "go".
        if self.num_enter == num_procs:
            self.num_enter = 0
            self.num_exit = 0
            self.flag = True
        self.lock.release()

        # Everybody waits until the flag is set to "go".
        while not self.flag:
            sleep(TICK)

        # Note that we exited.
        with self.lock:
            self.num_exit += 1
            if self.num_exit == num_procs:
                self.num_exit = -1



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""fake cifar 10 module."""
from argparse import ArgumentParser, Namespace

import numpy as np
from PIL import Image

from streaming.vision.convert.base import convert_image_class_dataset


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Args:
        Namespace: command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument('--out', type=str, default='/tmp/fake_cifar10/')
    args.add_argument('--num_train', type=int, default=8 * 8 * 16384)
    args.add_argument('--num_val', type=int, default=8 * 8 * 2048)
    return args.parse_args()


def make_split(root: str, split: str, count: int) -> None:
    """Process one split of the fake CIFAR10 dataset.

    Args:
        root (str): Output root directory.
        split (str): Dataset split name.
        count (int): Number of samples to create.
    """
    x = np.random.randint(0, 255, (count, 32, 32, 3), np.uint8)
    y = np.random.randint(0, 9, count)
    dataset = list(zip(map(Image.fromarray, x), y))
    convert_image_class_dataset(dataset=dataset, out_root=root, split=split)  # pyright: ignore


def main(args: Namespace) -> None:
    """Main: create streaming fake CIFAR10 dataset.

    Args:
        args (Namespace): command-line arguments.
    """
    make_split(args.out, 'train', args.num_train)
    make_split(args.out, 'val', args.num_val)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Register or look up the prefix to use for all shared resources.

The prefix is used by all workers using this StreamingDataset of this training job. This is used to
prevent shared resources like shared memory from colliding.
"""

import os
from collections import Counter
from tempfile import gettempdir
from time import sleep
from typing import Iterator, Union

import numpy as np
from torch import distributed as dist

from streaming.base.constant import BARRIER_FILELOCK, CACHE_FILELOCK, LOCALS, SHM_TO_CLEAN, TICK
from streaming.base.shared import SharedMemory
from streaming.base.world import World


def _each_prefix_int() -> Iterator[int]:
    """Get each possible prefix int to check in order.

    Returns:
        Iterator[int]: Each prefix int.
    """
    prefix_int = 0
    while True:
        yield prefix_int
        prefix_int += 1


def _get_path(prefix_int: int, name: str) -> str:
    """Get the name of the shared memory.

    Args:
        prefix (int): The prefix int.
        name (str): The name of the shared memory.

    Returns:
        str: Unique shared memory name.
    """
    return f'{prefix_int:06}_{name}'


def _pack_locals(dirnames: list[str], prefix_int: int) -> bytes:
    """Pack local dirnames and prefix int.

    Args:
        dirnames (List[str]): Unpacked local dirnames.
        prefix_int (int): Prefix int.

    Returns:
        bytes: Packed local dirnames and prefix int.
    """
    text = '\0'.join(dirnames) + f'\0{prefix_int}'
    data = text.encode('utf-8')
    size = 4 + len(data)
    return b''.join([np.int32(size).tobytes(), data])


def _unpack_locals(data: bytes) -> tuple[list[str], int]:
    """Unpack local dirnames and prefix int.

    Args:
        data (bytes): Packed local dirnames and prefix int.

    Returns:
        List[str]: Unpacked local dirnames and prefix int.
    """
    size = np.frombuffer(data[:4], np.int32)[0]
    text = data[4:size].decode('utf-8')
    text = text.split('\0')
    return text[:-1], int(text[-1] or 0)


def _check_self(streams_local: list[str]) -> None:
    """Check our local working directories for overlap.

    Args:
        streams_local (List[str]): Local dirs.

    Raises:
        ValueError: If there is overlap.
    """
    occurrences = Counter(streams_local)
    duplicate_local_dirs = [dirname for dirname, count in occurrences.items() if count > 1]
    if duplicate_local_dirs:
        raise ValueError(
            f'Reused local directory: {duplicate_local_dirs}. Provide a different one.')


def _check_and_find(streams_local: list[str], streams_remote: list[Union[str, None]],
                    shm_name: str) -> int:
    """Find the next available prefix while checking existing local dirs for overlap.

    Local leader walks the existing shm prefixes starting from zero, verifying that there is no
    local working directory overlap when remote directories exist. When attaching to an existing
    shm fails, we have reached the end of the existing shms. We will register the next one.

    Args:
        streams_local (List[str]): Our local working directories.
        streams_remote (List[Union[str, None]]): Our remote working directories.
        shm_name (str): The shared memory file name, e.g., LOCALS, BARRIER etc.

    Returns:
        int: Next available prefix int.
    """
    prefix_int = 0

    for prefix_int in _each_prefix_int():

        name = _get_path(prefix_int, shm_name)

        # Check if any shared memory filelocks exist for the current prefix
        try:
            filelock_exists = any(
                os.path.exists(os.path.join(gettempdir(), _get_path(prefix_int, filelock_name)))
                for filelock_name in [BARRIER_FILELOCK, CACHE_FILELOCK])
            if filelock_exists:
                continue
        except PermissionError:
            continue

        # Attempt to access shared memory by name. Use prefix_int if files do not exist
        try:
            shm = SharedMemory(name, False)
        except PermissionError:
            continue
        except FileNotFoundError:
            break

        if shm_name != LOCALS:
            continue

        their_locals, _ = _unpack_locals(bytes(shm.buf))

        # Do not check for a conflicting local directories across existing shared memory if
        # remote directories are None. Get the next prefix.
        if any(streams_remote):
            # Get the indices of the local directories which matches with the current
            # shared memory.
            matching_index = np.where(np.isin(streams_local, their_locals))[0]
            if matching_index.size > 0:
                for idx in matching_index:
                    # If there is a conflicting local directory for a non-None remote directory,
                    # raise an exception.
                    if streams_remote[idx] is not None:
                        raise ValueError(
                            f'Reused local directory: {streams_local} vs ' +
                            f'{their_locals}. Provide a different one. If using ' +
                            f'a unique local directory, try deleting the local directory and ' +
                            f'call `streaming.base.util.clean_stale_shared_memory()` only once ' +
                            f'in your script to clean up the stale shared memory before ' +
                            f'instantiation of `StreamingDataset`.')
    return prefix_int


def _check_and_find_retrying(streams_local: list[str], streams_remote: list[Union[str, None]],
                             shm_name: str, retry: int) -> int:
    """Find the next available prefix while checking existing dirs for overlap.

    If an overlap is found, sleeps for a tick and then tries again, up to "retry" times. We allow
    this grace period because modifying python shared memory in a destructor intermediated through
    a numpy array appears to be racy.

    Args:
        streams_local (List[str]): Our local working directories.
        streams_remote (List[Union[str, None]]): Our remote working directories.
        shm_name (str): The shared memory file name, e.g., LOCALS, BARRIER etc.
        retry (int): Number of retries upon failure before raising an exception.

    Returns:
        int: Next available prefix int.
    """
    if retry < 0:
        raise ValueError(f'Specify at least zero retries (provided {retry}).')
    errs = []
    for _ in range(1 + retry):
        try:
            return _check_and_find(streams_local, streams_remote, shm_name)
        except ValueError as err:
            errs.append(err)
            sleep(TICK)
    raise errs[-1]


def get_shm_prefix(streams_local: list[str],
                   streams_remote: list[Union[str, None]],
                   world: World,
                   retry: int = 100) -> tuple[int, SharedMemory]:
    """Register or lookup our shared memory prefix.

    Args:
        streams_local (List[str]): Local working dir of each stream, relative to /.
            We need to verify that there is no overlap with any other currently
            running StreamingDataset.
        streams_remote (List[Union[str, None]]): Remote working dir of each stream.
        world (World): Information about nodes, ranks, and workers.
        retry (int): Number of retries upon failure before raising an exception.
            Defaults to ``100``.

    Returns:
        Tuple[int, SharedMemory]: Shared memory integer prefix and object. The name
            is required to be very short due to limitations of Python on Mac OSX.
    """
    # Check my locals for overlap.
    _check_self(streams_local)

    prefix_int = max([
        _check_and_find_retrying(streams_local, streams_remote, shm_name=shm_name, retry=retry)
        for shm_name in SHM_TO_CLEAN
    ])

    if dist.is_available() and dist.is_initialized():
        dist.barrier()

    # First, the local leader registers the first available shm prefix, recording its locals.
    if world.is_local_leader:
        name = _get_path(prefix_int, LOCALS)
        data = _pack_locals(streams_local, prefix_int)
        shm = SharedMemory(name, True, len(data))
        shm.buf[:len(data)] = data

    if dist.is_available() and dist.is_initialized():
        dist.barrier()

    # Non-local leaders go next, searching for match.
    if not world.is_local_leader:
        name = _get_path(prefix_int, LOCALS)
        try:
            shm = SharedMemory(name, False)
        except FileNotFoundError:
            raise RuntimeError(f'Internal error: shared memory prefix={prefix_int} was not ' +
                               f'registered by local leader. This may be because you specified ' +
                               f'different ``local`` parameters from different ranks.')

        their_locals, their_prefix_int = _unpack_locals(bytes(shm.buf))
        if streams_local != their_locals or prefix_int != their_prefix_int:
            raise RuntimeError(f'Internal error: shared memory registered does not match ' +
                               f'local leader as streams_local or prefix_int not match. ' +
                               f'local leader: {their_locals} and {their_prefix_int}. ' +
                               f'expected: {streams_local} and {prefix_int}.')
    return prefix_int, shm  # pyright: ignore



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Determine shuffle quality of a run over a fixed number of samples."""

import logging

import numpy as np
from core.utils import remove_padded_samples
from numpy.typing import NDArray

from streaming.base.partition.orig import get_partitions_orig
from streaming.base.shuffle import get_shuffle

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def get_entropy(ordering: NDArray) -> float:
    """Calculate the entropy of an ordering, which is initially assumed to be in ascending order.

    Args:
        ordering (NDArray): The ordering to calculate the entropy of.

    Returns:
        float: The entropy of the ordering.
    """
    # get differences between elements
    diffs = np.diff(ordering)
    # diffs = np.insert(diffs, ordering.shape[0]-1, ordering[0]-ordering[-1])

    # change negative differences to positive
    diffs = np.abs(diffs)

    # count frequencies of differences
    diff_freqs = np.bincount(diffs)

    # remove zero frequency elements d
    diff_freqs = diff_freqs[diff_freqs != 0]

    # convert frequencies to probabilities
    diff_probs = diff_freqs / (ordering.shape[0] - 1)

    # calculate entropy
    diff_entropy = -np.sum(diff_probs * np.log2(diff_probs))

    return float(diff_entropy)


def get_partition_shard_info(epoch_size: int,
                             canonical_nodes: int,
                             physical_nodes: int,
                             devices: int,
                             workers: int,
                             device_batch_size: int,
                             samples_per_shard: int,
                             remove_padding: bool = False) -> tuple[NDArray, NDArray, NDArray]:
    """Partition up to 100 million samples and get associated shard information.

    Args:
        epoch_size (int): The number of samples in an epoch.
        canonical_nodes (int): The number of canonical nodes.
        physical_nodes (int): The number of physical nodes.
        devices (int): The number of devices.
        workers (int): The number of workers.
        device_batch_size (int): The batch size per device.
        samples_per_shard (int): Average number of samples per shard.
        remove_padding (bool): Whether to remove padding samples. Defaults to ``False``.

    Returns:
        Tuple[NDArray, NDArray, NDArray]: The partition, in order, the
            sizes of each shard, and the mapping of sample id to shard id.
    """
    num_samples = epoch_size
    if num_samples > 100_000_000:
        logger.warning(
            ' Epoch size is >100 million. Using 100 million samples to analyze shuffle quality.')
        num_samples = 100_000_000

    partition = get_partitions_orig(num_samples, canonical_nodes, physical_nodes, devices, workers,
                                    device_batch_size)
    partition = partition.transpose(3, 2, 0, 1, 4).flatten()
    if remove_padding:
        partition = remove_padded_samples(partition)

    # Construct shard sizes array.
    num_shards = num_samples // samples_per_shard
    shard_sizes = np.array([samples_per_shard] * num_shards)
    if num_samples % samples_per_shard != 0:
        num_shards += 1
        shard_sizes = np.append(shard_sizes, num_samples % samples_per_shard)

    # Construct sample id -> shard id mapping.
    shard_per_sample = np.repeat(np.arange(num_shards - 1), samples_per_shard)
    remaining_samples = num_samples - len(shard_per_sample)
    shard_per_sample = np.append(shard_per_sample, np.full(remaining_samples, num_shards - 1))

    return partition, shard_sizes, shard_per_sample


def get_entropy_shuffle_quality(shuffle_algo: str, partition: NDArray, shard_sizes: NDArray,
                                shard_per_sample: NDArray, canonical_nodes: int, seed: int,
                                shuffle_block_size: int) -> float:
    """Get the entropy of a shuffle, assuming samples and shards were initially in ascending order.

    Args:
        shuffle_algo (str): The shuffle algorithm to use.
        partition  (NDArray): The flattened, in-order partition to use.
        shard_sizes (NDArray): The sizes of each shard.
        shard_per_sample (NDArray): The mapping of sample id to shard id.
        canonical_nodes (int): The number of canonical nodes.
        seed (int): The seed to use for the shuffle.
        shuffle_block_size (int): The shuffle block size.

    Returns:
        float: The entropy metric, equal to the harmonic mean of sample and shard orderings.
    """
    if shuffle_algo != 'none':
        # Assume we are shuffling only for epoch 0.
        shuffle_ordering = get_shuffle(shuffle_algo, shard_sizes, canonical_nodes, seed, 0,
                                       shuffle_block_size)
        partition = shuffle_ordering[partition]
    sample_entropy = get_entropy(partition)
    shard_entropy = get_entropy(shard_per_sample[partition])
    return (2 * sample_entropy * shard_entropy) / (sample_entropy + shard_entropy)


def analyze_shuffle_quality_entropy(algo: str, canonical_nodes: int, physical_nodes: int,
                                    devices: int, workers: int, device_batch_size: int,
                                    shuffle_block_size: int, samples_per_shard: int,
                                    epoch_size: int, seed: int) -> tuple[str, float]:
    """Analyze the quality of a shuffle for one algorithm.

    Args:
        algo (str): The algorithm to analyze.
        canonical_nodes (int): The number of canonical nodes.
        physical_nodes (int): The number of physical nodes.
        devices (int): The number of devices.
        workers (int): The number of workers.
        device_batch_size (int): The batch size per device.
        shuffle_block_size (int): The shuffle block size.
        samples_per_shard (int): Average number of samples per shard.
        epoch_size (int): The number of samples in an epoch.
        seed (int): The seed to use for the shuffle.

    Returns:
        Tuple[str, float]: Shuffle algorithm and shuffle quality.
    """
    logger.info(f' Analyzing shuffle quality for {algo}...')

    # Getting partition, shard_sizes, and shard_per_sample only has to be done once for all algos.
    partition, shard_sizes, shard_per_sample = get_partition_shard_info(epoch_size,
                                                                        canonical_nodes,
                                                                        physical_nodes,
                                                                        devices,
                                                                        workers,
                                                                        device_batch_size,
                                                                        samples_per_shard,
                                                                        remove_padding=True)

    shuffle_quality = get_entropy_shuffle_quality(algo, partition, shard_sizes, shard_per_sample,
                                                  canonical_nodes, seed, shuffle_block_size)

    return algo, shuffle_quality



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""A dataset, or sub-dataset if mixing, from which we stream/cache samples."""

import hashlib
import json
import os
import tempfile
from typing import Optional, Sequence

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Self

from streaming.base.compression import decompress
from streaming.base.constant import TICK
from streaming.base.distributed import barrier, get_local_rank
from streaming.base.format import FileInfo, Reader, get_index_basename, reader_from_json
from streaming.base.hashing import get_hash
from streaming.base.storage import download_file
from streaming.base.util import retry, wait_for_file_to_exist
from streaming.base.world import World


class Stream:
    """A dataset, or sub-dataset if mixing, from which we stream/cache samples.

    We initialize a StreamingDataset with one or more Streams. Streams may be resampled to achieve
    different mixtures of samples.

    Stream init takes three kinds of arguments:

    * At least one of ``remote`` and ``local`` must exist. If no ``remote``, the data must be
      local. If no ``local``, we cache to a temp directory.

      * ``remote``
      * ``local``

    * At most one of ``proportion``, ``repeat``, or ``choose`` may exist. If provided one of these,
      we derive the rest. Note that ``proportion`` (relative) and ``repeat``/``choose`` (absolute)
      are mutually incompatible -- you must entirely use one or the other (or neither) for all
      sub-datasets. If none are provided for all streams and ``epoch_size`` is unspecified, then
      each sample from each stream is seen once per epoch. If none are provided for all streams
      and ``epoch_size`` is specified, then streams are sampled in proportion to their size.

      * ``proportion``
      * ``repeat``
      * ``choose``

    * The remaining arguments are optional knobs for controlling downloading behavior and default
      to ``None``. If ``None``, they take a default value provided to or by the StreamingDataset
      init.

      * ``split``
      * ``download_retry``
      * ``download_timeout``
      * ``validate_hash``
      * ``keep_zip``

    Args:
        remote (str, optional): Remote path or directory to download the dataset from. If ``None``,
            its data must exist locally. Defaults to ``None``.
        local (str, optional): Local working directory to download shards to. This is where shards
            are cached while they are being used. Uses a temp directory if not set. Defaults to
            ``None``.
        split (str, optional): Which dataset split to use, if any. If provided, we stream from/to
            the ``split`` subdirs of  ``remote`` and ``local``. Defaults to ``None``.
        proportion (float, optional): How much to upsample or downsample this sub-dataset, as the
            proportion of the total combined dataset that consists of this sub-dataset. If
            using proportions, all sub-datasets provided together to the StreamingDataset init must
            define their proportions. The total combined number of samples is either the
            StreamingDataset argument "epoch_size" if provided, or kept the same total size as the
            underlying data if not. If provided, must be non-negative. Defaults to ``None``.
        repeat (float, optional): How much to upsample or downsample this sub-dataset, as a
            multipler on the number of samples. If provided, must be non-negative. Defaults to
            ``None``.
        choose (int, optional): How much to upsample or downsample this sub-dataset, as the exact
            number of resulting samples. If provided, must be non-negative. Defaults to ``None``.
        download_retry (int, optional): Number of download re-attempts before giving up. Defaults
            to ``None``.
        download_timeout (float, optional): Number of seconds to wait for a shard to download
            before raising an exception. Defaults to ``None``.
        validate_hash (str, optional): Optional hash or checksum algorithm to use to validate
            shards. Defaults to ``None``.
        keep_zip (bool, optional): Whether to keep or delete the compressed form when decompressing
            downloaded shards. If ``False``, keep if and only if remote is local or no remote.
            Defaults to ``None``.
    """

    def __init__(self,
                 *,
                 remote: Optional[str] = None,
                 local: Optional[str] = None,
                 split: Optional[str] = None,
                 proportion: Optional[float] = None,
                 repeat: Optional[float] = None,
                 choose: Optional[int] = None,
                 download_retry: Optional[int] = None,
                 download_timeout: Optional[float] = None,
                 validate_hash: Optional[str] = None,
                 keep_zip: Optional[bool] = None) -> None:
        self.remote = remote
        self._local = local
        self.split = split or ''

        has_proportion = proportion is not None
        has_repeat = repeat is not None
        has_choose = choose is not None
        if not (0 <= has_proportion + has_repeat + has_choose <= 1):
            raise ValueError('At most one of `proportion`, `repeat`, and `choose` may be ' +
                             'specified; the others are derived')

        self._proportion = proportion
        if proportion is not None:
            if proportion < 0:
                raise ValueError('`proportion` must be non-negative')
            self.proportion = proportion

        self._repeat = repeat
        if repeat is not None:
            if repeat < 0:
                raise ValueError('`repeat` must be non-negative')
            self.repeat = repeat

        self._choose = choose
        if choose is not None:
            if choose < 0:
                raise ValueError('`choose` must be non-negative')
            self.choose = choose

        self._download_retry = download_retry
        if download_retry is not None:
            if download_retry < 0:
                raise ValueError('`download_retry` must be non-negative')
            self.download_retry = download_retry

        self._download_timeout = download_timeout
        if download_timeout is not None:
            if download_timeout <= 0:
                raise ValueError('`download_timeout` must be positive')
            self.download_timeout = download_timeout

        self.validate_hash = validate_hash

        if local is None:
            self.local = self._get_temporary_directory()
            if get_local_rank() == 0:
                if os.path.exists(self.local):
                    raise FileExistsError(
                        f'Could not create a temporary local directory {self.local} because it ' +
                        f'already exists. If you want to reuse the locally cached dataset, ' +
                        f'explicitly pass in a unique local directory with the `local` argument.' +
                        f' Otherwise, delete this directory and retry.')
                os.makedirs(self.local)
            barrier()
        else:
            self.local = local

        self._keep_zip = keep_zip
        if keep_zip is not None:
            self.keep_zip = keep_zip
            self.safe_keep_zip = self.keep_zip or self.remote in {None, self.local}

    def _get_temporary_directory(self) -> str:
        """Construct a path to a temporary directory based on remote and split."""
        root = tempfile.gettempdir()
        hash = ''
        if self.remote is not None:
            hash = hashlib.blake2s(self.remote.encode('utf-8'), digest_size=16).hexdigest()
        return os.path.join(root, hash, self.split)

    def apply_default(self, default: dict) -> None:
        """Apply defaults, setting any unset fields.

        We use pairs of (name, _name) in order to make type checking happy.

        Args:
            default (Self): Stream containing default values for all optional fields.
        """
        if not (self.remote or self._local):
            raise ValueError('`remote` and/or `local` path must be provided')

        if not self.split:
            self.split = default['split'] or ''
        if self._download_retry is None:
            self.download_retry = default['download_retry']
        if self._download_timeout is None:
            self.download_timeout = default['download_timeout']
        if self.validate_hash is None:
            self.validate_hash = default['validate_hash'] or None
        if self._keep_zip is None:
            self.keep_zip = default['keep_zip']
            self.safe_keep_zip = default['keep_zip'] or self.remote in {None, self.local}

    @classmethod
    def validate_weights(cls, streams: Sequence[Self]) -> tuple[bool, bool]:
        """Validate stream weights, returning whether relative or absolute weighting was used.

        Args:
            streams (Sequence[Stream]): Every stream comprising the dataset.

        Returns:
            bool: Whether streams are weighted relatively (proportionally).
        """
        # Validate stream weights ("proportion", "repeat", "choose", or none).
        is_proportional = hasattr(streams[0], 'proportion')
        is_unspecified = True
        for stream_id, stream in enumerate(streams):
            has_proportion = hasattr(stream, 'proportion')
            has_repeat = hasattr(stream, 'repeat')
            has_choose = hasattr(stream, 'choose')
            if not (0 <= has_proportion + has_repeat + has_choose <= 1):
                raise ValueError(f'Streams must provide at most one of `proportion`, `repeat`, ' +
                                 f'or `choose` (error in stream {stream_id})')
            if is_proportional != has_proportion:
                raise ValueError(f'Relative (`proportion`) and absolute (`repeat`, `choose`, ' +
                                 f'none) stream weights are incompatible with each other (error ' +
                                 f'in stream {stream_id})')
            if has_proportion or has_repeat or has_choose:
                is_unspecified = False
        return is_proportional, is_unspecified

    @classmethod
    def apply_weights(cls, streams: Sequence[Self], samples_per_stream: NDArray[np.int64],
                      choose_per_epoch: Optional[int], seed: int) -> int:
        """Given samples per stream, derive each stream's proportion/repeat/samples.

        Modifies streams to save the derived weights.

        Args:
            streams (Sequence[Stream]): The list of streams which comprise the dataset.
            samples_per_stream (NDArray[np.int64]): Underlying samples of each stream.
            choose_per_epoch (int, optional): Absolute epoch size if weighting relatively.
            seed (int): Random number generator seed used to sample evenly.

        Returns:
            int: Number of samples to draw per epoch.
        """
        # Validate provided weights, determining whether they are relative or absolute.
        are_weights_relative, are_weights_unspecified = cls.validate_weights(streams)

        # Derive weights.
        if are_weights_relative:
            # Relative.
            if not choose_per_epoch:
                choose_per_epoch = sum(samples_per_stream)
            proportion_per_stream = np.array([stream.proportion for stream in streams], np.float64)
            proportion_per_stream /= proportion_per_stream.sum()
            choose_per_stream = (choose_per_epoch * proportion_per_stream).astype(np.int64)
            shortfall = choose_per_epoch - choose_per_stream.sum()
            rng = np.random.default_rng(seed)
            indices = rng.choice(len(streams), shortfall, False)
            choose_per_stream[indices] += 1
            repeat_per_stream = choose_per_stream / samples_per_stream
        elif are_weights_unspecified and choose_per_epoch:
            # weights are unspecified, but epoch size (choose_per_epoch) is provided.
            # sample from each stream in proportion stream's samples
            proportion_per_stream = samples_per_stream.copy().astype(np.float64)
            proportion_per_stream /= proportion_per_stream.sum()
            choose_per_stream = (choose_per_epoch * proportion_per_stream).astype(np.int64)
            shortfall = choose_per_epoch - choose_per_stream.sum()
            rng = np.random.default_rng(seed)
            indices = rng.choice(len(streams), shortfall, False)
            choose_per_stream[indices] += 1
            repeat_per_stream = choose_per_stream / samples_per_stream
        else:
            # Absolute.
            if choose_per_epoch:
                raise ValueError('Only provide `choose` when weighting streams relatively')
            choose_per_stream = np.zeros(len(streams), np.int64)
            for stream_id, stream in enumerate(streams):
                if hasattr(stream, 'repeat'):
                    choose = int(stream.repeat * samples_per_stream[stream_id])
                elif hasattr(stream, 'choose'):
                    choose = stream.choose
                else:
                    choose = samples_per_stream[stream_id]
                choose_per_stream[stream_id] = choose
            repeat_per_stream = choose_per_stream / samples_per_stream
            proportion_per_stream = choose_per_stream / choose_per_stream.sum()
            choose_per_epoch = sum(choose_per_stream)

        # Now that we know the true props/reps/choices, inject those back into the streams.
        for stream, proportion, repeat, choose in zip(streams, proportion_per_stream,
                                                      repeat_per_stream, choose_per_stream):
            stream.proportion = proportion
            stream.repeat = repeat
            stream.choose = choose

        return int(choose_per_epoch)

    def _download_file(self, from_basename: str, to_basename: Optional[str] = None) -> str:
        """Safely download a file from remote to local cache.

        Args:
            from_basename (str): Source basename.
            to_basename (str, optional): Destination basename, if different.

        Returns:
            str: Local cache filename.
        """
        # Calculate paths.
        if self.remote is None:
            remote = None
        else:
            remote = os.path.join(self.remote, self.split, from_basename)
        local = os.path.join(self.local, self.split, to_basename or from_basename)

        # Attempt to download, possibly repeating on failure.
        retry(num_attempts=self.download_retry)(
            lambda: download_file(remote, local, self.download_timeout))()

        return local

    def _decompress_shard_part(self, zip_info: FileInfo, zip_filename: str, raw_filename: str,
                               compression: Optional[str]) -> None:
        """Validate and decompress shard data.

        Args:
            zip_info (FileInfo): Compressed file info.
            zip_filename (str): Compressed filename.
            raw_filename (str): Decompressed filename.
            compression (str, optional): Compression algorithm.
        """
        # Load compressed.
        data = open(zip_filename, 'rb').read()

        # Validate what was downloaded.
        if self.validate_hash:
            if self.validate_hash not in zip_info.hashes:
                raise ValueError(
                    f'Hash algorithm `{self.validate_hash}` chosen for data ' +
                    f'validation does not match with those provided during dataset ' +
                    f'creation `{sorted(zip_info.hashes.keys())}`. Provide one of those.')
            if get_hash(self.validate_hash, data) != zip_info.hashes[self.validate_hash]:
                raise ValueError(f'Checksum failure: {zip_filename}')

        # Decompress and save that.
        data = decompress(compression, data)  # pyright: ignore
        tmp_filename = raw_filename + '.tmp'
        with open(tmp_filename, 'wb') as out:
            out.write(data)
        os.rename(tmp_filename, raw_filename)

        # Maybe remove compressed to save space.
        if not self.safe_keep_zip:
            os.remove(zip_filename)

    def _prepare_shard_part(self,
                            raw_info: FileInfo,
                            zip_info: Optional[FileInfo] = None,
                            compression: Optional[str] = None) -> int:
        """Get shard data given metadata for the raw and compressed versions of it.

        MDS format uses joint shards (ie, one file per shard). Other formats supported by streaming
        use split shards (ie, shard data lives in two files per shard: the raw data itself and
        metadata in a separate file).

        Args:
            raw_info (FileInfo): Raw file info.
            zip_info (FileInfo, optional): Zip file info. Defaults to ``None``.
            compression (str, optional): Compression algorithm used for zip_info. Defaults to
                ``None``.

        Returns:
            int: Change in cache usage.
        """
        # Has raw?
        delta = 0
        raw_filename = os.path.join(self.local, self.split, raw_info.basename)
        if os.path.isfile(raw_filename):
            # Has raw.
            if zip_info and not self.safe_keep_zip:
                zip_filename = os.path.join(self.local, self.split, zip_info.basename)
                if os.path.isfile(zip_filename):
                    # If don't keep zip and it has a zip, drop the zip.
                    os.remove(zip_filename)
                    delta -= zip_info.bytes
        else:
            # Missing raw. Uses zip?
            if zip_info:
                # Ensure has zip.
                zip_filename = os.path.join(self.local, self.split, zip_info.basename)
                if not os.path.isfile(zip_filename):
                    self._download_file(zip_info.basename)
                    delta += zip_info.bytes

                # Validate and decompress.
                self._decompress_shard_part(zip_info, zip_filename, raw_filename, compression)
                delta += raw_info.bytes
                if not self.safe_keep_zip:
                    delta -= zip_info.bytes
            else:
                # Download raw.
                self._download_file(raw_info.basename)
                delta += raw_info.bytes

                # Validate.
                if self.validate_hash:
                    if self.validate_hash not in raw_info.hashes:
                        raise ValueError(
                            f'Hash algorithm `{self.validate_hash}` chosen for data ' +
                            f'validation does not match with those provided during dataset ' +
                            f'creation `{sorted(raw_info.hashes.keys())}`. Provide one of those.')
                    data = open(raw_filename, 'rb').read()
                    if get_hash(self.validate_hash, data) != raw_info.hashes[self.validate_hash]:
                        raise ValueError(f'Checksum failure: {raw_filename}')
        return delta

    def prepare_shard(self, shard: Reader) -> int:
        """Ensure (download, validate, extract, etc.) that we have the given shard.

        Args:
            shard (Reader): Which shard.

        Returns:
            int: Change in cache usage.
        """
        delta = 0
        for raw_info, zip_info in shard.file_pairs:
            delta += self._prepare_shard_part(raw_info, zip_info, shard.compression)
        return delta

    def get_shards(self, world: World, allow_unsafe_types: bool) -> list[Reader]:
        """Load this Stream's index, retrieving its shard readers.

        Args:
            world (World): Distributed context.
            allow_unsafe_types (bool): If a shard contains Pickle, which allows arbitrary code
                execution during deserialization, whether to keep going if ``True`` or raise an
                error.

        Returns:
            `List[Reader]: Shard readers.
        """
        # Download the index file if it does not exist locally.
        basename = get_index_basename()
        filename = os.path.join(self.local, self.split, basename)  # pyright: ignore
        if not os.path.exists(filename):
            if world.is_local_leader:
                if self.remote:
                    # Downloads the `index.json` as `index.json.tmp` fully and then rename it to
                    # `index.json` since only one process downloads the `index.json` file while
                    # other processes wait for it to get downloaded. Hence, It avoids loading the
                    # in-progress downloading `index.json`.
                    tmp_filename = self._download_file(basename, basename + '.tmp')
                    os.rename(tmp_filename, filename)
                else:
                    if not os.path.exists(filename):
                        raise RuntimeError(f'No `remote` provided, but local file {filename} ' +
                                           'does not exist either')
            else:
                wait_for_file_to_exist(
                    filename, TICK, self.download_timeout,
                    f'Index file {os.path.join(self.remote or "", self.split or "", basename)} ' +
                    f'-> {filename} took too long to download or failed to download. Either increase the '
                    + f'`download_timeout` value or check the local rank 0 traceback.')

        # Load the index.
        try:
            obj = json.load(open(filename))
        except json.decoder.JSONDecodeError as error:
            error.args = (f'Index file at {filename} is empty or corrupted. ' + error.args[0],)
            raise error

        # Version check.
        if obj['version'] != 2:
            raise ValueError(f'Unsupported streaming data version: {obj["version"]}. ' +
                             f'Expected version 2.')

        # Initialize shard readers according to the loaded info.
        shards = []
        for info in obj['shards']:
            shard = reader_from_json(self.local, self.split, info)
            shard.validate(allow_unsafe_types)
            shards.append(shard)

        return shards

    def set_up_local(self, shards: list[Reader], cache_usage_per_shard: NDArray[np.int64]) -> None:
        """Bring a local directory into a consistent state, getting which shards are present.

        Args:
            shards (List[Reader]): List of this stream's shards.
            cache_usage_per_shard (NDArray[np.int64]): Cache usage per shard of this stream.
        """
        # List the cache directory (so that we hit the filesystem once).
        local_dirname = os.path.join(self.local, self.split)
        listing = set()
        for dirname, _, subfiles in os.walk(local_dirname):
            for subfile in subfiles:
                filename = os.path.join(dirname, subfile)
                listing.add(filename)

        # Determine which shards are present, making local dir consistent.
        for i, shard in enumerate(shards):
            cache_usage_per_shard[i] = shard.set_up_local(listing, self.safe_keep_zip)

    def get_index_size(self) -> int:
        """Get the size of the index file in bytes.

        Returns:
            int: Size in bytes.
        """
        filename = os.path.join(self.local, self.split, get_index_basename())
        return os.stat(filename).st_size



from argparse import ArgumentParser
from glob import glob
from multiprocessing import Pool
import os
import tensorflow.compat.v1 as tf


def parse_args():
    args = ArgumentParser()
    args.add_argument('--dataset', type=str, required=True)
    return args.parse_args()


def get_num_samples(filename):
    dataset = tf.data.TFRecordDataset(filename)
    count = 0
    for _ in dataset:
        count += 1
    return count


def main(args):
    pattern = os.path.join(args.dataset, '*')
    filenames = sorted(glob(pattern))
    total = 0
    with Pool() as pool:
        it = pool.imap(get_num_samples, filenames)
        for filename, count in zip(filenames, it):
            print(filename, count)
            total += count
    print(total)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Contains info about the nodes, ranks, and workers of the run for simulation purposes."""

from streaming.base.world import World


class SimulationWorld(World):
    """Contains info about the nodes, ranks, and workers of the run, for simulation.

    Nodes are all assumed to contain the same number of devices (via local_world_size).

    Nodes:
      - node / num_nodes
      - is_multinode

    Ranks:
      - rank / num_ranks
      - rank_of_node / ranks_per_node

    Workers:
      - worker / num_workers
      - worker_of_node / workers_per_node
      - worker_of_rank / workers_per_rank
      - is_leader
      - is_local_leader
    Args:
        nodes (int): The number of nodes.
        devices (int): The number of devices per node.
        workers (int): The number of workers per device.
    """

    def __init__(self, nodes: int, devices: int, workers: int):

        # For simulation purposes, we take in the nodes, devices, and workers from the
        # SimulationDataset, and assume we are always rank 0 and worker 0.
        self.rank = 0
        self.num_ranks = nodes * devices
        self.ranks_per_node = devices
        self.rank_of_node = self.rank % self.ranks_per_node
        self.node = self.rank // self.ranks_per_node
        self.num_nodes = self.num_ranks // self.ranks_per_node
        self.is_multinode = 1 < self.num_nodes

        self.worker_of_rank = 0
        self.workers_per_rank = workers

        self.worker = self.rank * self.workers_per_rank + self.worker_of_rank
        self.num_workers = self.num_ranks * self.workers_per_rank
        self.worker_of_node = self.rank_of_node * self.workers_per_rank + self.worker_of_rank
        self.workers_per_node = self.ranks_per_node * self.workers_per_rank

        self.is_leader = not self.worker
        self.is_local_leader = not self.worker_of_node



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Objects that live in shared memory.

For when using `threading` or `multiprocessing` from the python standard library won't do, because
we are coordinating separately instantiated pytorch worker processes.
"""

from streaming.base.shared.array import SharedArray as SharedArray
from streaming.base.shared.barrier import SharedBarrier as SharedBarrier
from streaming.base.shared.memory import SharedMemory as SharedMemory
from streaming.base.shared.prefix import _get_path as _get_path
from streaming.base.shared.prefix import get_shm_prefix as get_shm_prefix
from streaming.base.shared.scalar import SharedScalar as SharedScalar

__all__ = ['SharedArray', 'SharedBarrier', 'SharedMemory', 'get_shm_prefix', 'SharedScalar']



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Streaming package setup."""

import os

import setuptools
from setuptools import setup

# Read the streaming version
# Cannot import from `streaming.__version__` since that will not be available when building or installing the package
with open(os.path.join(os.path.dirname(__file__), 'streaming', '_version.py')) as f:
    version_globals = {}
    version_locals = {}
    exec(f.read(), version_globals, version_locals)
    streaming_version = version_locals['__version__']

# Use repo README for PyPi description
with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

# Hide the content between <!-- SETUPTOOLS_LONG_DESCRIPTION_HIDE_BEGIN --> and
# <!-- SETUPTOOLS_LONG_DESCRIPTION_HIDE_END --> tags in the README
while True:
    start_tag = '<!-- SETUPTOOLS_LONG_DESCRIPTION_HIDE_BEGIN -->'
    end_tag = '<!-- SETUPTOOLS_LONG_DESCRIPTION_HIDE_END -->'
    start = long_description.find(start_tag)
    end = long_description.find(end_tag)
    if start == -1:
        assert end == -1, 'there should be a balanced number of start and ends'
        break
    else:
        assert end != -1, 'there should be a balanced number of start and ends'
        long_description = long_description[:start] + long_description[end + len(end_tag):]

classifiers = [
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
]

install_requires = [
    'boto3>=1.21.45,<2',
    'Brotli>=1.0.9',
    'google-cloud-storage>=2.9.0,<2.11.0',
    'matplotlib>=3.5.2,<4',
    'paramiko>=2.11.0,<4',
    'python-snappy>=0.6.1,<1',
    'torch>=1.10,<3',
    'torchvision>=0.10',
    'tqdm>=4.64.0,<5',
    'transformers>=4.21.3,<5',
    'xxhash>=3.0.0,<4',
    'zstd>=1.5.2.5,<2',
    'oci>=2.88,<3',
    'azure-storage-blob>=12.0.0,<13',
    'azure-storage-file-datalake>=12.11.0,<13',
    'azure-identity>=1.13.0',
]

extra_deps = {}

extra_deps['dev'] = [
    'datasets>=2.4.0,<4',
    'pyarrow>14.0.0',
    'docformatter>=1.4',
    'jupyter==1.1.1',
    'pre-commit>=2.18.1,<4',
    'pytest==8.3.3',
    'pytest_codeblocks==0.17.0',
    'pytest-cov>=4,<6',
    'toml==0.10.2',
    'yamllint==1.35.1',
    'moto>=4.0,<6',
    'fastapi==0.115.4',
    'pydantic==2.9.2',
    'uvicorn==0.32.0',
    'pytest-split==0.10.0',
]

extra_deps['docs'] = [
    'GitPython==3.1.42',
    'docutils==0.17.1',
    'furo==2022.9.29',
    'myst-parser==0.16.1',
    'nbsphinx==0.9.1',
    'pandoc==2.3',
    'pypandoc==1.13',
    'sphinx-argparse==0.4.0',
    'sphinx-copybutton==0.5.2',
    'sphinx==4.4.0',
    'sphinx-tabs==3.4.5',
    'sphinxcontrib.katex==0.9.6',
    'sphinxcontrib-applehelp==1.0.0',
    'sphinxcontrib-devhelp==1.0.0',
    'sphinxcontrib-htmlhelp==2.0.0',
    'sphinxcontrib-qthelp==1.0.0',
    'sphinxcontrib-serializinghtml==1.1.5',
]

extra_deps['simulator'] = [
    'sortedcollections>=2.1.0,<3',
    'streamlit>=1.26.0,<2',
    'altair>=5.1.1,<6',
    'omegaconf>=2.3.0,<3',
    'PyYAML>=6.0,<7',
    'pandas>=2.0.3,<3',
    'wandb>=0.15.5,<1',
    'humanize>=4.7.0,<5',
]

extra_deps['spark'] = [
    'pyspark>=3,<4',
]

extra_deps['databricks'] = [
    'databricks-sdk==0.29.0',
]

extra_deps['alipan'] = [
    'AliPCS-Py>=0.8,<1',
]

extra_deps['hf'] = [
    'huggingface_hub>=0.23.4,<0.27',
]

extra_deps['testing'] = [
    'mosaicml-cli>=0.5.25,<0.7',
]

extra_deps['all'] = sorted({dep for deps in extra_deps.values() for dep in deps})

package_name = os.environ.get('MOSAIC_PACKAGE_NAME', 'mosaicml-streaming')

if package_name != 'mosaicml-streaming':
    print(f'Building mosaicml-streaming as {package_name}')

setup(
    name=package_name,
    version=streaming_version,
    author='MosaicML',
    author_email='team@mosaicml.com',
    description=
    'Streaming lets users create PyTorch compatible datasets that can be streamed from cloud-based object stores',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mosaicml/streaming/',
    include_package_data=True,
    package_data={
        'streaming': ['py.typed'],
    },
    packages=setuptools.find_packages(exclude=['tests*']),
    entry_points={
        'console_scripts': ['simulator = simulation.launcher:launch_simulation_ui',],
    },
    classifiers=classifiers,
    install_requires=install_requires,
    extras_require=extra_deps,
    python_requires='>=3.9',
)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Near replica of StreamingDataset for simulation purposes."""

import logging
import os
import shutil
import time
import warnings
from math import ceil
from typing import Optional, Sequence, Union

import numpy as np
from core.sim_spanner import SimulationSpanner
from core.sim_world import SimulationWorld
from numpy.typing import NDArray

from streaming.base import Stream, StreamingDataset
from streaming.base.batching import generate_work
from streaming.base.format import get_index_basename
from streaming.base.spanner import Spanner
from streaming.base.util import bytes_to_int, number_abbrev_to_int

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class SimulationDataset(StreamingDataset):
    """Near replica of StreamingDataset for simulation purposes.

    Args:
        nodes (int): Number of nodes.
        devices (int): Number of devices.
        workers (int): Number of workers.
        streams (Optional[Sequence[Stream]]): One or more streams to stream/cache samples from,
            which may be upsampled or downsampled. StreamingDataset uses either ``streams`` or
            ``remote``/``local``. Defaults to ``None``.
        remote (Optional[str]): Remote path or directory to download the dataset from. If ``None``,
            its data must exist locally. StreamingDataset uses either ``streams`` or
            ``remote``/``local``. Defaults to ``None``.
        local (Optional[str]): Local working directory to download shards to. This is where shards
            are cached while they are being used. Uses a temp directory if not set.
            StreamingDataset uses either ``streams`` or ``remote``/``local``. Defaults to ``None``.
        split (Optional[str]): Which dataset split to use, if any. If provided, we stream from/to
            the ``split`` subdirs of  ``remote`` and ``local``. Defaults to ``None``.
        download_retry (int): Number of download re-attempts before giving up. Defaults to ``2``.
        download_timeout (float): Number of seconds to wait for a shard to download before raising
            an exception. Defaults to ``60``.
        validate_hash (Optional[str]): Optional hash or checksum algorithm to use to validate
            shards. Defaults to ``None``.
        keep_zip (bool): Whether to keep or delete the compressed form when decompressing
            downloaded shards. If ``False``, keep iff remote is local or no remote. Defaults to
            ``False``.
        epoch_size (Union[int, str], optional): Number of samples to draw per epoch balanced across all
            streams. If ``None``, takes its value from the total number of underlying samples.
            Provide this field if you are weighting streams relatively to target a larger or
            smaller epoch size. Defaults to ``None``. Can also take in human-readable number
            abbreviations (e.g., ``"100k"``, ``"64M"``, ``"77b"``, and so on). Defaults to ``None``.
        predownload (int, optional): Target number of samples to download per worker in advance
            of current sample. Workers will attempt to download ahead by this many samples during,
            but not before, training. Recommendation is to provide a value greater than per device
            batch size to ensure at-least per device batch size number of samples cached locally.
            If ``None``, its value is set to ``8 * batch_size``. Defaults to ``None``.
        cache_limit (Union[int, str], optional): Maximum size in bytes of this StreamingDataset's
            shard cache. Before downloading a shard, the least recently used resident shard(s)
            may be evicted (deleted from the local cache) in order to stay under the limit.
            Set to ``None`` to disable shard eviction. Supports integer bytes as well as string
            human-readable bytes (e.g., ``100b``, ``64kb``, ``77mb``, and so on). Defaults to
            ``None``.
        partition_algo (str): Which partitioning algorithm to use. Defaults to ``relaxed``.
        num_canonical_nodes (int, optional): Canonical number of nodes for shuffling with
            resumption. The sample space is divided evenly according to the number of canonical
            nodes. The higher the value, the more independent non-overlapping paths the
            StreamingDataset replicas take through the shards per model replica (increasing data
            source diversity). If ``None``, this is interpreted as 64 times the number of physical
            nodes of the initial run if ``shuffle_algo`` is ``py1s`` or ``py2s``, and simply the
            number of physical nodes of the initial run otherwise. Defaults to ``None``.

            .. note::

                For sequential sample ordering, set ``shuffle`` to ``False`` and
                ``num_canonical_nodes`` to the number of physical nodes of the initial run.
        batch_size (int, optional): Per-device batch size, the same as what is passed to the
            DataLoader. This affects how the dataset is partitioned over the workers and is
            necessary for deterministic resumption and optimal performance. Defaults to ``None``.
        shuffle (bool): Whether to iterate over the samples in randomized order. Defaults to
            ``False``.
        shuffle_algo (str): Which shuffling algorithm to use. Defaults to ``py1e``.
        shuffle_seed (int): Seed for Deterministic data shuffling. Defaults to ``9176``.
        shuffle_block_size (int, optional): Unit of shuffle. A canonical node's samples are split
            into blocks of this size, and samples within each block are shuffled. If ``None``, its
            value is calculated as ``max(4_000_000 // num_canonical_nodes), 1 << 18)``. Defaults to
            ``None``.
        sampling_method (str): Which sampling method to use, either ``balanced`` or ``fixed``.
            Defaults to ``balanced``.
        sampling_granularity (int): When picking samples for a stream's final partial repeat,
            how many samples to pick from the same shard at a time (``1`` for evenly balanced
            across shards, ``1000`` to pick 1000 samples from the same shard at a time, etc).
            Defaults to ``1``.
        batching_method (str): Which batching method to use, either ``random``, ``stratified``, or
            ``per_stream``. Defaults to ``random``.
        allow_unsafe_types (bool): If a shard contains Pickle, which allows arbitrary code
            execution during deserialization, whether to keep going if ``True`` or raise an error
            if ``False``. Defaults to ``False``.
    """

    def __init__(self,
                 nodes: int,
                 devices: int,
                 workers: int,
                 streams: Optional[Sequence[Stream]] = None,
                 remote: Optional[str] = None,
                 local: Optional[str] = None,
                 split: Optional[str] = None,
                 download_retry: int = 2,
                 download_timeout: float = 60,
                 validate_hash: Optional[str] = None,
                 keep_zip: bool = False,
                 epoch_size: Optional[Union[int, str]] = None,
                 predownload: Optional[int] = None,
                 cache_limit: Optional[Union[int, str]] = None,
                 partition_algo: str = 'relaxed',
                 num_canonical_nodes: Optional[int] = None,
                 batch_size: Optional[int] = None,
                 shuffle: bool = False,
                 shuffle_algo: str = 'py1e',
                 shuffle_seed: int = 9176,
                 shuffle_block_size: Optional[int] = None,
                 sampling_method: str = 'balanced',
                 sampling_granularity: int = 1,
                 batching_method: str = 'random',
                 allow_unsafe_types: bool = False) -> None:

        # Time how long it takes for StreamingDataset instantiation
        t0 = time.time()

        # Global arguments (which do not live in Streams).
        self.nodes = nodes
        self.devices = devices
        self.workers = workers
        self.cache_limit = cache_limit
        self.partition_algo = partition_algo
        self.predownload = predownload
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.shuffle_algo = shuffle_algo
        self.shuffle_seed = shuffle_seed
        self.shuffle_block_size = shuffle_block_size
        self.sampling_method = sampling_method
        self.sampling_granularity = sampling_granularity
        self.batching_method = batching_method
        self.num_canonical_nodes = num_canonical_nodes
        self.allow_unsafe_types = allow_unsafe_types

        self.initial_physical_nodes = nodes

        # Set num_canonical_nodes based on the shuffling algorithm chosen.
        if self.num_canonical_nodes is None:
            if self.shuffle_algo in ['py1s', 'py2s']:
                self.num_canonical_nodes = 64 * self.nodes
            else:
                self.num_canonical_nodes = self.nodes

        # Set shuffle_block_size if not provided, based on num_canonical_nodes.
        if self.shuffle_block_size is None:
            self.shuffle_block_size = max(4_000_000 // self.num_canonical_nodes, 1 << 18)

        # Check streams vs remote/local.
        if bool(streams) == (bool(remote) or bool(local)):
            raise ValueError(
                'You must provide either `streams` or `remote`/`local`, but not both.')

        # Check sampling method is one of "balanced" or "fixed".
        if self.sampling_method not in ['balanced', 'fixed']:
            raise ValueError(
                f'Invalid sampling method: {sampling_method}. Must be one of `balanced` or `fixed`.'
            )

        # Check sampling method is one of "balanced" or "fixed".
        if self.batching_method not in ['random', 'per_stream', 'stratified']:
            raise ValueError(
                f'Invalid batching method: {batching_method}. Must be one of `random`, \
                    `per_stream`, or `stratified`.')

        # Check that predownload is at least per device batch size, and set it if currently `None`.
        if self.predownload is not None and self.batch_size is not None and \
            self.predownload < self.batch_size:
            warnings.warn(f'predownload < batch_size ({self.predownload} < {self.batch_size}).' +
                          f'This may result in slower batch time. Recommendation is to set ' +
                          f'predownload to at-least batch_size.')
        elif self.predownload is None:
            self.predownload = 8 * self.batch_size if self.batch_size is not None else 64

        self.batch_size = batch_size

        # Convert epoch size from string to int, if needed. Cannot be negative.
        epoch_size_value = None
        if epoch_size:
            epoch_size_value = number_abbrev_to_int(epoch_size)
            if epoch_size_value < 0:
                raise ValueError(f'Epoch size cannot be negative. Received {epoch_size_value}.')

        # Initialize the Stream defaults and normalize to a list of Streams.
        if streams:
            default = {
                'remote': remote,
                'local': local,
                'split': split,
                'download_retry': download_retry,
                'download_timeout': download_timeout,
                'validate_hash': validate_hash,
                'keep_zip': keep_zip,
            }
            for stream in streams:
                stream.apply_default(default)
        else:
            default = Stream(remote=remote,
                             local=local,
                             split=split,
                             download_retry=download_retry,
                             download_timeout=download_timeout,
                             validate_hash=validate_hash,
                             keep_zip=keep_zip)
            streams = [default]

        # Validate the stream weighting scheme (relative or absolute) to catch errors before we go
        # to the trouble of loading them.
        Stream.validate_weights(streams)

        # Set streams.
        self.streams = streams
        self.num_streams = len(streams)

        self.stream_info = {}
        # 0 means index file is remote, 1 means local, 2 means created
        indices_created = []
        for stream_idx, stream in enumerate(self.streams):
            if stream.remote:
                filepath = os.path.join(stream.remote, stream.split, get_index_basename())
                indices_created.append(0)
            else:
                filepath = os.path.join(stream.local, stream.split, get_index_basename())
                # This suffix means a mock index file was created. Have to clean up later.
                if stream.local.split('_')[-1] == 'indexcreated':
                    indices_created.append(2)
                else:
                    # Index file is local. Don't delete later.
                    indices_created.append(1)
            self.stream_info[stream_idx] = {
                'path': filepath,
                'local': stream.local,
                'remote': stream.remote,
                'proportion': stream._proportion,
                'repeat': stream._repeat,
                'choose': stream._choose
            }

        # Initialize the SimulationWorld, which tells us about nodes/devices/workers
        self.world = SimulationWorld(self.nodes, self.devices, self.workers)

        # Download each stream's index, load their shards, and map streams <-> shards.
        self.num_samples = 0
        self.shards = []
        stream_per_shard = []
        self.shard_offset_per_stream = np.zeros(self.num_streams, np.int64)
        self.shards_per_stream = np.zeros(self.num_streams, np.int64)
        self.sample_offset_per_stream = np.zeros(self.num_streams, np.int64)
        self.samples_per_stream = np.zeros(self.num_streams, np.int64)
        index_filenames = []
        local_foldernames = []
        for stream_id, stream in enumerate(self.streams):
            logger.info(f' Processing index file for stream {stream_id + 1}')
            stream_shards = stream.get_shards(self.world, self.allow_unsafe_types)
            num_stream_samples = sum(map(len, stream_shards))
            index_filename = os.path.join(stream.local, stream.split, get_index_basename())
            index_filenames.append(index_filename)
            local_foldernames.append(stream.local)
            if not num_stream_samples:
                raise RuntimeError(f'Stream contains no samples: {index_filename}.')
            stream_per_shard += [stream_id] * len(stream_shards)
            self.shard_offset_per_stream[stream_id] = len(self.shards)
            self.shards_per_stream[stream_id] = len(stream_shards)
            self.sample_offset_per_stream[stream_id] = self.num_samples
            self.samples_per_stream[stream_id] = num_stream_samples
            self.shards += stream_shards
            self.num_samples += num_stream_samples

        self.stream_per_shard = np.array(stream_per_shard, np.int64)
        self.num_shards = len(self.shards)

        # Check that cache limit is possible.
        if self.cache_limit:
            if isinstance(self.cache_limit, str):
                self.cache_limit = bytes_to_int(self.cache_limit)
            min_cache_usage = sum((stream.get_index_size() for stream in streams))
            if self.cache_limit <= min_cache_usage:
                raise ValueError(f'Minimum cache usage ({min_cache_usage} bytes) is larger than ' +
                                 f'the cache limit ({self.cache_limit} bytes). Please raise ' +
                                 f'`cache_limit`.')

        for stream_idx, index_filename in enumerate(index_filenames):
            if indices_created[stream_idx] == 0:
                # Index file was downloaded from remote.
                try:
                    # Remove the index.json file.
                    os.remove(index_filename)
                except FileNotFoundError:
                    pass
            elif indices_created[stream_idx] == 1:
                # Index file was local. Don't delete.
                pass
            else:
                # Directory and index file were created. Delete both.
                shutil.rmtree(local_foldernames[stream_idx])

        # Build the shard index (for partitioning and mapping samples to shards).
        self.samples_per_shard = np.array([shard.samples for shard in self.shards], np.int64)
        self.sample_offset_per_shard = self.samples_per_shard.cumsum() - self.samples_per_shard
        self.spanner = SimulationSpanner(self.samples_per_shard)

        # Also keep track of the raw and compressed sizes of each shard, indexed by shard_id.
        self.raw_shard_sizes = np.array([shard.get_raw_size() for shard in self.shards], np.int64)
        self.zip_shard_sizes = np.array([shard.get_zip_size() or 0 for shard in self.shards],
                                        np.int64)

        logger.info(f' Total number of shards: {self.num_shards}')
        logger.info(f' Average number of samples per shard: {self.num_samples / self.num_shards}')
        logger.info(f' Average raw shard size (bytes): {np.mean(self.raw_shard_sizes)}')
        logger.info(f' Average zip shard size (bytes): {np.mean(self.zip_shard_sizes)}')

        # Now that we know the number of underlying samples of each stream, derive each stream's
        # true proportion/repeat/choose, as well as the total epoch size.
        self.epoch_size = Stream.apply_weights(self.streams, self.samples_per_stream,
                                               epoch_size_value, self.shuffle_seed)

        # Length (__len__) is the resampled epoch size divided over the number of devices.
        self.length = ceil(self.epoch_size / self.world.num_ranks)

        t1 = time.time()
        self.instantiation_time = t1 - t0

        logger.info(' SimulationDataset created successfully.')

    def get_sample_partition(self, epoch: int, sample_in_epoch: int) -> NDArray:
        """Get the dataset's partition of this epoch's sample space.

        Args:
            epoch (int): Which epoch it is.
            sample_in_epoch (int): Where we are in the epoch.

        Returns:
            NDArray[np.int64]: Our partition of the epoch.
        """
        return generate_work(self.batching_method, self, self.world, epoch, sample_in_epoch)

    def get_samples_per_node(self, epoch: int, sample_in_epoch: int) -> NDArray:
        """Get the dataset's number of samples per node, worker, device.

        Args:
            epoch (int): Which epoch it is.
            sample_in_epoch (int): Where we are in the epoch.

        Returns:
            NDArray[np.int64]: The dataset's samples per node, worker, device.
        """
        partition = generate_work(self.batching_method, self, self.world, epoch, sample_in_epoch)
        # Modify partition to be in traversal order, per node, device, and worker.
        return partition.reshape(self.nodes, self.devices, self.workers, -1)

    def get_spanner(self) -> Spanner:
        """Get the dataset's spanner object, which does global sample id indexing.

        Returns:
            Spanner: The dataset's spanner object.
        """
        return self.spanner

    def get_raw_shard_sizes(self) -> NDArray[np.int64]:
        """Get the dataset's raw shard sizes.

        Returns:
            NDArray[np.int64]: The dataset's raw shard sizes.
        """
        return self.raw_shard_sizes

    def get_zip_shard_sizes(self) -> NDArray[np.int64]:
        """Get the dataset's zip shard sizes.

        Returns:
            NDArray[np.int64]: The dataset's zip shard sizes.
        """
        return self.zip_shard_sizes

    def get_nodes(self) -> int:
        """Get the dataset's number of nodes.

        Returns:
            int: The dataset's number of nodes.
        """
        return self.nodes

    def get_devices(self) -> int:
        """Get the dataset's number of devices.

        Returns:
            int: The dataset's number of devices.
        """
        return self.devices

    def get_workers(self) -> int:
        """Get the dataset's number of workers.

        Returns:
            int: The dataset's number of workers.
        """
        return self.workers

    def get_num_canonical_nodes(self) -> int:
        """Get the dataset's number of canonical nodes.

        Returns:
            int: The dataset's number of canonical nodes.
        """
        if not isinstance(self.num_canonical_nodes, int):
            raise TypeError(f'`self.num_canonical_nodes` must be an int. ' +
                            f'Got {type(self.num_canonical_nodes)} instead.')
        return self.num_canonical_nodes

    def get_batch_size(self) -> int:
        """Get the dataset's batch size.

        Returns:
            int: The dataset's batch size.
        """
        if not isinstance(self.batch_size, int):
            raise TypeError(f'`self.batch_size` must be an int. ' +
                            f'Got {type(self.batch_size)} instead.')
        return self.batch_size

    def get_num_shards(self) -> int:
        """Get the dataset's number of shards.

        Returns:
            int: The dataset's number of shards.
        """
        return self.num_shards

    def get_avg_samples_per_shard(self) -> int:
        """Get the dataset's average number of samples per shard.

        Returns:
            int: The dataset's average number of samples per shard.
        """
        return round(self.num_samples / self.num_shards)

    def get_predownload(self) -> int:
        """Get the dataset's predownload.

        Returns:
            int: The dataset's predownload.
        """
        if not isinstance(self.predownload, int):
            raise TypeError(f'`self.predownload` must be an int. ' +
                            f'Got {type(self.predownload)} instead.')
        return self.predownload

    def get_cache_limit(self) -> Optional[int]:
        """Get the dataset's cache limit.

        Returns:
            Optional[int]: The dataset's cache limit.
        """
        if isinstance(self.cache_limit, str):
            self.cache_limit = bytes_to_int(self.cache_limit)
        return self.cache_limit

    def get_instantiation_time(self) -> float:
        """Get the dataset's instantiation time.

        Returns:
            float: The dataset's instantiation time.
        """
        return self.instantiation_time

    def get_num_batches(self) -> int:
        """Get the dataset's number of batches.

        Returns:
            int: The dataset's number of batches.
        """
        if self.batch_size is None:
            raise ValueError(f'Cannot get number of batches without `batch size`, had ' +
                             f'`batch_size` of `None`')
        return self.epoch_size // (self.batch_size * self.devices * self.nodes)

    def get_stream_info(self) -> dict:
        """Get the dataset's stream info.

        Returns:
            dict: The dataset's stream info.
        """
        return self.stream_info

    def get_shuffle(self) -> bool:
        """Get the dataset's shuffle.

        Returns:
            bool: The dataset's shuffle.
        """
        return self.shuffle

    def get_shuffle_algo(self) -> str:
        """Get the dataset's shuffle algorithm.

        Returns:
            str: The dataset's shuffle algorithm.
        """
        return self.shuffle_algo

    def get_shuffle_seed(self) -> int:
        """Get the dataset's shuffle seed.

        Returns:
            int: The dataset's shuffle seed.
        """
        return self.shuffle_seed

    def get_shuffle_block_size(self) -> int:
        """Get the dataset's shuffle block size.

        Returns:
            int: The dataset's shuffle block size.
        """
        if not isinstance(self.shuffle_block_size, int):
            raise TypeError(f'`self.shuffle_block_size` must be an int. ' +
                            f'Got {type(self.shuffle_block_size)} instead.')
        return self.shuffle_block_size

    def get_epoch_size(self) -> int:
        """Get the dataset's epoch size.

        Returns:
            int: The dataset's epoch size.
        """
        return self.epoch_size

    def get_sampling_method(self) -> str:
        """Get the dataset's sampling method.

        Returns:
            str: The dataset's sampling method.
        """
        return self.sampling_method

    def get_sampling_granularity(self) -> int:
        """Get the dataset's sampling granularity.

        Returns:
            int: The dataset's sampling granularity.
        """
        return self.sampling_granularity

    def get_batching_method(self) -> str:
        """Get the dataset's batching method.

        Returns:
            str: The dataset's batching method.
        """
        return self.batching_method



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Apportion shards/samples to nodes/ranks/workers for elastically deterministic sample order."""

import logging
import math
import warnings
from typing import Optional

import numpy as np
from numpy.typing import NDArray

logger = logging.getLogger(__name__)


def get_partitions_orig(num_samples: int,
                        num_canonical_nodes: int,
                        num_physical_nodes: int,
                        ranks_per_node: int,
                        workers_per_rank: int,
                        batch_size: int,
                        drop_first: int = 0,
                        initial_physical_nodes: Optional[int] = None) -> NDArray[np.int64]:
    """Partition the given number of samples to nodes, ranks, and workers.

    Either canonical or physical nodes must be evenly divisible by the other.

    It is suggested to set num_canonical_nodes higher than your expected number of physical nodes,
    because scaling your number of nodes below that level may result in more shards being used
    across node boundaries due to preserving the same global sample order.

    Args:
        num_samples (int): Dataset size.
        num_canonical_nodes (int): Number of canonical nodes.
        num_physical_nodes (int): Number of physical nodes.
        ranks_per_node (int): Number of ranks per node.
        workers_per_rank (int): Number of worker partitions per rank.
        batch_size (int): Batch size of DataLoader and dataset, which affects how the dataset is
            partitioned over the workers.
        drop_first (int): Number of samples seen already, which are dropped. Defaults to ``0``.
        initial_physical_nodes (int, optional): Number of physical nodes at the start of training.
            Defaults to ``None``.

    Returns:
        NDArray[np.int64]: Partitions of shape (physical nodes, ranks per node, workers per rank,
            batches per worker, batch size).
    """
    if num_canonical_nodes < num_physical_nodes:
        if num_physical_nodes % num_canonical_nodes:
            raise ValueError('Either canonical or physical nodes must be evenly divisible by ' +
                             'the other, otherwise striping slices of shards over nodes may ' +
                             'lead to each node downloading all shards')
    elif num_physical_nodes < num_canonical_nodes:
        if num_canonical_nodes % num_physical_nodes:
            raise ValueError('Either canonical or physical nodes must be evenly divisible by ' +
                             'the other, otherwise striping slices of shards over nodes may ' +
                             'lead to each node downloading all shards')

    # If drop_first isn't a multiple of num_physical_nodes, round down to make it divisible.
    if drop_first % num_physical_nodes:
        logger.warning(
            '`drop_first` was not divisible by `num_physical_nodes`. Rounding it down ' +
            'to make it divisible.')
        drop_first -= drop_first % num_physical_nodes

    # Divide the full dataset sample range into a sample range per canonical node.
    samples_per_canonical_node = (num_samples + num_canonical_nodes - 1) // num_canonical_nodes
    node_ratio = 0
    padding = 0
    if num_canonical_nodes < num_physical_nodes:
        node_ratio = num_physical_nodes // num_canonical_nodes
        overflow = samples_per_canonical_node % node_ratio
        if overflow:
            padding = node_ratio - overflow
    padded_samples_per_canonical_node = samples_per_canonical_node + padding

    # For samples to be properly split across canonical nodes, there must be more samples than nodes.
    # The edge case is when the number of samples is equal to the number of canonical nodes, but this only works when
    # there is an equal or greater number of canonical nodes than physical nodes.
    # If these conditions are not met, an alternative sampling approach is used that leads to many repeats.
    if num_samples > num_canonical_nodes or (num_samples == num_canonical_nodes and
                                             num_canonical_nodes >= num_physical_nodes):
        # Create the initial sample ID matrix.
        #
        # ids: (canonical nodes, padded samples per canonical node).
        ids = np.arange(num_canonical_nodes * padded_samples_per_canonical_node, dtype=np.int64)
        ids = ids.reshape(num_canonical_nodes, padded_samples_per_canonical_node)

        # Adjust row offsets to ignore the padding.
        #
        # row_offsets: (canonical nodes, 1).
        row_offsets = np.arange(num_canonical_nodes) * padding
        row_offsets = np.expand_dims(row_offsets, 1)
        ids -= row_offsets

        # Reconfigure where each row starts iterating for irregular-sized rows.
        #
        # row_starts: (canonical nodes, 1).
        row_starts = np.arange(num_canonical_nodes) * num_samples // num_canonical_nodes
        row_starts = np.expand_dims(row_starts, 1)
        ids += row_starts - ids[:, :1]

        # For short rows (length not evenly divisible), repeat the last ID to get even length.
        #
        # row_stops: (canonical nodes, 1).
        row_stops = np.arange(1, 1 + num_canonical_nodes) * num_samples // num_canonical_nodes
        row_stops = np.expand_dims(row_stops, 1)
        are_rows_short = row_stops - row_starts < samples_per_canonical_node
        ids[:, samples_per_canonical_node - 1:samples_per_canonical_node] -= are_rows_short

        # If padding we needed, repeat samples to populate it.
        if padding:
            ids[:, -padding:] = ids[:,
                                    -padding - node_ratio + 1 - padding:-padding - node_ratio + 1]
    else:
        warnings.warn(f'Trying to partition {num_samples} samples over {num_canonical_nodes}' +
                      f' canonical nodes. This will result in many samples being repeated, and ' +
                      f'depending on your batching method, batches being completely dropped. ' +
                      f'Check if your dataset has the expected number of samples, and consider ' +
                      f'decreasing the number of canonical nodes.')
        shape_needed = (num_canonical_nodes, padded_samples_per_canonical_node)
        total_samples_needed = num_canonical_nodes * padded_samples_per_canonical_node
        current_samples = np.arange(num_samples, dtype=np.int64)
        full_repeats = total_samples_needed // num_samples
        leftover_samples = total_samples_needed % num_samples
        ids = np.concatenate(
            [np.tile(current_samples, full_repeats), current_samples[:leftover_samples]])
        ids = ids.reshape(shape_needed)

    # Flatten, drop samples that have already been seen, reshape back.
    #
    # ids: (physical nodes, samples per node).
    ids = ids.transpose()
    ids = ids.flatten()
    ids = ids[drop_first:]
    ids = ids.reshape(-1, num_physical_nodes)
    ids = ids.transpose()

    # Interleave the node sample ranges over each node's ranks, padding with -1 for reshaping.
    #
    # ids: (physical nodes, samples per rank, ranks per node).
    overflow = ids.shape[1] % ranks_per_node
    if overflow:
        underflow = ranks_per_node - overflow
        enough_padding_samples = ranks_per_node + underflow - 1 <= ids.shape[1]
        if enough_padding_samples:
            last = ids[:, -ranks_per_node - underflow + 1:-ranks_per_node + 1]
        else:
            # There are less samples than ranks. Usually, we pad by trying to ensure that the same
            # samples don't get repeated over and over, but with in this case, we are forced to.
            warnings.warn(f'Attempting to partition {ids.shape[1]} samples per physical node ' +
                          f'over {ranks_per_node} gpus. This will result in many samples being ' +
                          f'repeated, and depending on your batching method, batches being ' +
                          f'completely dropped. Check if your dataset has the expected number ' +
                          f'of samples.')
            num_samples = ids.shape[1]
            full_repeats = underflow // num_samples
            leftover_samples = underflow % num_samples
            last = np.concatenate([np.tile(ids, full_repeats), ids[:, :leftover_samples]], 1)
        ids = np.concatenate([ids, last], 1)

    ids = ids.reshape(num_physical_nodes, -1, ranks_per_node)

    # Pad with -1 adequately for reshaping across workers.
    #
    # ids: (physical nodes, samples per rank, ranks per node).
    overflow = ids.shape[1] % workers_per_rank
    rounded_num_samples = math.ceil(
        ids.shape[1] / (workers_per_rank * batch_size)) * (workers_per_rank * batch_size)
    overflow = rounded_num_samples - ids.shape[1]
    if overflow:
        last = np.full((num_physical_nodes, overflow, ranks_per_node), -1, np.int64)
        ids = np.concatenate([ids, last], 1)

    # Interleave each rank's padded samples across its workers.
    #
    # ids: (physical nodes, ranks per node, workers per rank, batches per worker, batch size).
    ids = ids.transpose(0, 2, 1)
    ids = ids.reshape(num_physical_nodes, ranks_per_node, -1, workers_per_rank, batch_size)
    return ids.transpose(0, 1, 3, 2, 4)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Peripheral functions for simulation functionality."""

import numpy as np
from core.sim_dataset import SimulationDataset
from core.sim_time import Time, TimeUnit
from numpy.typing import NDArray


def get_batches_epochs(dataset: SimulationDataset, max_duration: Time) -> tuple[int, int, int]:
    """Get batches per epoch, epochs, and total epochs from a Time object.

    Args:
        dataset (SimulationDataset): The dataset being simulated.
        max_duration (Time): The maximum duration, can be specified in yaml.

    Returns:
        Tuple[int, int, int]: batches per epoch, epochs, and the total batches.
    """
    # get epochs, batches_per_epoch, and total_batches from a Time obect
    dataset_batches = dataset.get_num_batches()
    batches_per_epoch = dataset_batches
    epochs = 1
    total_batches = dataset_batches
    if max_duration.unit == TimeUnit.EPOCH:
        epochs = max_duration.value
        batches_per_epoch = dataset_batches
        total_batches = epochs * batches_per_epoch
    elif max_duration.unit == TimeUnit.BATCH:
        full_epochs = max_duration.value // dataset_batches
        # check if there is a partial epoch we should fulfill
        if max_duration.value % dataset_batches != 0:
            full_epochs += 1
        # make sure we don't simulate past the duration set.
        if max_duration.value < dataset_batches:
            batches_per_epoch = max_duration.value
        else:
            batches_per_epoch = dataset_batches
        total_batches = max_duration.value
    else:
        raise ValueError('Simulator currently only supports max_duration in epochs or batches.')

    return batches_per_epoch, epochs, total_batches


def get_total_batches(dataset: SimulationDataset, max_duration: Time) -> int:
    """Get total batches from a Time object.

    Args:
        dataset (SimulationDataset): The dataset being simulated.
        max_duration (Time): The maximum duration, can be specified in yaml.

    Returns:
        int: The total batches.
    """
    dataset_batches = dataset.get_num_batches()
    total_batches = dataset_batches
    if max_duration.unit == TimeUnit.EPOCH:
        epochs = max_duration.value
        batches_per_epoch = dataset_batches
        total_batches = epochs * batches_per_epoch
    elif max_duration.unit == TimeUnit.BATCH:
        total_batches = max_duration.value
    else:
        raise ValueError('Simulator currently only supports max_duration in epochs or batches.')

    return total_batches


def remove_padded_samples(samples: NDArray) -> NDArray:
    """Remove padded samples from a batch.

    Args:
        samples (NDArray): The samples to remove padded samples from.

    Returns:
        NDArray: The samples with padded samples removed.
    """
    return np.delete(samples, np.where(samples == -1))


def bytes_to_time(bytes: int, bandwidth: int) -> float:
    """Convert bytes to time.

    Args:
        bytes (int): The bytes to convert.
        bandwidth (int): The bandwidth available.

    Returns:
        float: The time it takes to transfer the bytes.
    """
    return bytes / bandwidth


def time_to_bytes(time: float, bandwidth: int) -> int:
    """Convert time to bytes.

    Args:
        time (float): The time to convert.
        bandwidth (int): The bandwidth available.

    Returns:
        int: The bytes transferred in the time.
    """
    return int(time * bandwidth)


def get_rolling_avg_throughput(step_times: NDArray, window: int = 10) -> NDArray:
    """Get rolling average throughput from step times.

    Args:
        step_times (NDArray): time per step, as calculated by simulation
        window (int): window size for rolling average

    Returns:
        NDArray: rolling average throughput
    """
    step_times_rolling_avg = np.convolve(step_times, np.ones(window) / window, mode='valid')
    batch_throughput_rolling_avg = 1 / step_times_rolling_avg
    batch_throughput_rolling_avg = np.concatenate(
        (np.array([0] * (window - 1)), batch_throughput_rolling_avg))

    return batch_throughput_rolling_avg


def get_simulation_stats(step_times: NDArray, time_per_sample: float,
                         device_batch_size: int) -> tuple[int, float, int, int]:
    """Gets simulation stats for web UI.

    Args:
        step_times (NDArray): time per step, as calculated by simulation
        time_per_sample (float): time to process one sample on one device (seconds)
        device_batch_size (int): batch size per device

    Returns:
        Tuple[int, float, int, int]: number of steps with throughput drops, time till warmup,
            step number of warmup, number of steps with throughput drops after warmup
    """
    # calculate percent of download-limited steps
    min_step_time = time_per_sample * device_batch_size
    all_throughput_drops = int(np.count_nonzero(step_times > (min_step_time)))

    epsilon = 1e-6

    # calculate warmup time (time to first max possible rolling average throughput) within epsilon
    max_throughput = 1 / min_step_time
    rolling_avg_throughput = get_rolling_avg_throughput(step_times)
    if np.max(rolling_avg_throughput) >= max_throughput - epsilon:
        warmup_step = int(np.argmax(rolling_avg_throughput >= (max_throughput)) + 1)
        warmup_time = float(np.sum(step_times[:warmup_step]))
    else:
        # we never hit the max possible throughput
        warmup_step = int(rolling_avg_throughput.shape[0])
        warmup_time = float(np.sum(step_times))

    # see if there are throughput drops after warmup so we can notify users
    if warmup_step != rolling_avg_throughput.shape[0]:
        # if we did hit the max throughput then we check for later drops
        post_warmup_tp_drops = int(np.count_nonzero(step_times[warmup_step:] > min_step_time))
    else:
        # since warmup was the whole time, there are no post-warmup throughput drops
        post_warmup_tp_drops = 0

    return all_throughput_drops, warmup_time, warmup_step, post_warmup_tp_drops



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Serialize samples into streaming dataset shards and index."""

import json
import logging
import os
import shutil
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures._base import Future
from threading import Event
from time import sleep
from types import TracebackType
from typing import Any, Optional, Union

from typing_extensions import Self

from streaming.base.compression import compress, get_compression_extension, is_compression
from streaming.base.format.index import get_index_basename
from streaming.base.hashing import get_hash, is_hash
from streaming.base.storage.upload import CloudUploader
from streaming.base.util import bytes_to_int

__all__ = ['JointWriter', 'SplitWriter']

logger = logging.getLogger(__name__)


class Writer(ABC):
    """Writes a streaming dataset.

    Args:
        out (str | Tuple[str, str]): Output dataset directory to save shard files.

            1. If ``out`` is a local directory, shard files are saved locally.
            2. If ``out`` is a remote directory, a local temporary directory is created to
               cache the shard files and then the shard files are uploaded to a remote
               location. At the end, the temp directory is deleted once shards are uploaded.
            3. If ``out`` is a tuple of ``(local_dir, remote_dir)``, shard files are saved in the
               `local_dir` and also uploaded to a remote location.
        keep_local (bool): If the dataset is uploaded, whether to keep the local dataset directory
            or remove it after uploading. Defaults to ``False``.
        compression (str, optional): Optional compression or compression:level. Defaults to
            ``None``.
        hashes (List[str], optional): Optional list of hash algorithms to apply to shard files.
            Defaults to ``None``.
        size_limit (Union[int, str], optional): Optional shard size limit, after which point
            to start a new shard. If ``None``, puts everything in one shard. Can specify bytes
            human-readable format as well, for example ``"100kb"`` for 100 kilobyte
            (100*1024) and so on. Defaults to ``1 << 26``.
        extra_bytes_per_shard (int): Extra bytes per serialized shard (for computing shard size
            while writing). Defaults to ``0``.
        extra_bytes_per_sample (int): Extra bytes per serialized sample (for computing shard size
            while writing). Defaults to ``0``.
        **kwargs (Any): Additional settings for the Writer.

            progress_bar (bool): Display TQDM progress bars for uploading output dataset files to
                a remote location. Default to ``False``.
            max_workers (int): Maximum number of threads used to upload output dataset files in
                parallel to a remote location. One thread is responsible for uploading one shard
                file to a remote location. Default to ``min(32, (os.cpu_count() or 1) + 4)``.
            retry (int): Number of times to retry uploading a file to a remote location.
                Default to ``2``.
            exist_ok (bool): If the local directory exists and is not empty, whether to overwrite
                the content or raise an error. `False` raises an error. `True` deletes the
                content and starts fresh. Defaults to `False`.
    """

    format: str = ''  # Name of the format (like "mds", "csv", "json", etc).

    def __init__(self,
                 *,
                 out: Union[str, tuple[str, str]],
                 keep_local: bool = False,
                 compression: Optional[str] = None,
                 hashes: Optional[list[str]] = None,
                 size_limit: Optional[Union[int, str]] = 1 << 26,
                 extra_bytes_per_shard: int = 0,
                 extra_bytes_per_sample: int = 0,
                 **kwargs: Any) -> None:

        compression = compression or None
        if compression:
            if not is_compression(compression):
                raise ValueError(f'Invalid compression: {compression}.')

        hashes = hashes or []
        if list(hashes) != sorted(hashes):
            raise ValueError('Hashes must be unique and in sorted order.')
        for algo in hashes:
            if not is_hash(algo):
                raise ValueError(f'Invalid hash: {algo}.')

        size_limit_value = None
        if size_limit:
            size_limit_value = bytes_to_int(size_limit)
            if size_limit_value < 0:
                raise ValueError(f'`size_limit` must be greater than zero, instead, ' +
                                 f'found as {size_limit_value}.')
            if size_limit_value >= 2**32:
                raise ValueError(f'`size_limit` must be less than 2**32, instead, ' +
                                 f'found as {size_limit_value}. This is because sample ' +
                                 f'byte offsets are stored with uint32.')

        # Validate keyword arguments
        invalid_kwargs = [
            arg for arg in kwargs.keys()
            if arg not in ('progress_bar', 'max_workers', 'retry', 'exist_ok')
        ]
        if invalid_kwargs:
            raise ValueError(f'Invalid Writer argument(s): {invalid_kwargs} ')

        self.keep_local = keep_local
        self.compression = compression
        self.hashes = hashes
        self.size_limit = size_limit_value
        self.extra_bytes_per_shard = extra_bytes_per_shard
        self.extra_bytes_per_sample = extra_bytes_per_sample
        self.new_samples: list[bytes]
        self.new_shard_size: int

        self.shards = []

        # Remove local directory if requested prior to creating writer
        local = os.path.expanduser(out) if isinstance(out, str) else os.path.expanduser(out[0])
        if os.path.exists(local) and kwargs.get('exist_ok', False):
            logger.warning(
                f'Directory {local} exists and is not empty; exist_ok is set to True so will remove contents.'
            )
            shutil.rmtree(local)
        self.cloud_writer = CloudUploader.get(out, keep_local, kwargs.get('progress_bar', False),
                                              kwargs.get('retry', 2))
        self.local = self.cloud_writer.local
        self.remote = self.cloud_writer.remote
        # `max_workers`: The maximum number of threads that can be executed in parallel.
        # One thread is responsible for uploading one shard file to a remote location.
        self.executor = ThreadPoolExecutor(max_workers=kwargs.get('max_workers', None))
        # Create an event to track an exception in a thread.
        self.event = Event()

        self._reset_cache()

    def _reset_cache(self) -> None:
        """Reset our internal shard-building cache.

        This is called on init or after writing a shard.
        """
        self.new_samples = []
        self.new_shard_size = self.extra_bytes_per_shard

    @abstractmethod
    def encode_sample(self, sample: dict[str, Any]) -> bytes:
        """Encode a sample dict to bytes.

        Args:
            sample (Dict[str, Any]): Sample dict.

        Returns:
            bytes: Sample encoded as bytes.
        """
        raise NotImplementedError

    def _name_next_shard(self, extension: Optional[str] = None) -> tuple[str, Optional[str]]:
        """Get the filenames of the next shard to be created.

        Args:
            extension (str): Optional additional extension (eg, meta files).

        Returns:
            Tuple[str, str]: Pair of (decompressed, compressed) filenames.
        """
        shard = len(self.shards)
        parts = ['shard', f'{shard:05}', self.format]
        if extension:
            parts.append(extension)
        raw_basename = '.'.join(parts)
        if self.compression:
            ext = get_compression_extension(self.compression)
            parts.append(ext)
            zip_basename = '.'.join(parts)
        else:
            zip_basename = None
        return raw_basename, zip_basename

    def _hash(self, data: bytes, basename: str) -> dict[str, Any]:
        """Generate file metadata.

        Args:
            data (bytes): The file data.
            basename (str): The file's basename.

        Returns:
            Dict[str, Any]: File metadata.
        """
        hashes = {}
        for algo in self.hashes:
            hashes[algo] = get_hash(algo, data)
        return {'basename': basename, 'bytes': len(data), 'hashes': hashes}

    def _process_file(self, raw_data: bytes, raw_basename: str,
                      zip_basename: Optional[str]) -> tuple[dict, Optional[dict]]:
        """Process and save a shard file (hash, compress, hash, write).

        Args:
            raw_data (bytes): Uncompressed data.
            raw_basename (str): Uncompressed basename.
            zip_basename (str): Compressed basename.

        Returns:
            Dict[str, Any]: Metadata containing basename, size, and hashes.
        """
        raw_info = self._hash(raw_data, raw_basename)
        if zip_basename:
            zip_data = compress(self.compression, raw_data)
            zip_info = self._hash(zip_data, zip_basename)
            data = zip_data
            basename = zip_basename
        else:
            zip_info = None
            data = raw_data
            basename = raw_basename
        filename = os.path.join(self.local, basename)
        with open(filename, 'wb') as out:
            out.write(data)
        return raw_info, zip_info

    def get_config(self) -> dict[str, Any]:
        """Get object describing shard-writing configuration.

        Returns:
            Dict[str, Any]: JSON object.
        """
        return {
            'version': 2,
            'format': self.format,
            'compression': self.compression,
            'hashes': self.hashes,
            'size_limit': self.size_limit
        }

    @abstractmethod
    def flush_shard(self) -> None:
        """Flush cached samples to storage, creating a new shard."""
        raise NotImplementedError

    def write(self, sample: dict[str, Any]) -> None:
        """Write a sample.

        May flush an entire new shard, then caches the sample.

        Args:
            sample (Dict[str, Any]): Sample dict.
        """
        if self.event.is_set():
            # Shutdown the executor and cancel all the pending futures due to exception in one of
            # the threads.
            self.cancel_future_jobs()
            raise Exception('One of the threads failed. Check other traceback for more ' +
                            'details.')
        # Execute the task if there is no exception in any of the async threads.
        new_sample = self.encode_sample(sample)
        new_sample_size = len(new_sample) + self.extra_bytes_per_sample
        if self.size_limit and self.size_limit < self.new_shard_size + new_sample_size:
            self.flush_shard()
            self._reset_cache()
        self.new_samples.append(new_sample)
        self.new_shard_size += new_sample_size

    def _write_index(self) -> None:
        """Write the index, having written all the shards."""
        if self.new_samples:
            raise RuntimeError('Internal error: not all samples have been written.')
        basename = get_index_basename()
        filename = os.path.join(self.local, basename)
        obj = {
            'version': 2,
            'shards': self.shards,
        }
        with open(filename, 'w') as out:
            json.dump(obj, out, sort_keys=True)
        # Execute the task if there is no exception in any of the async threads.
        while self.executor._work_queue.qsize() > 0:
            logger.debug(
                f'Queue size: {self.executor._work_queue.qsize()}. Waiting for all ' +
                f'shard files to get uploaded to {self.remote} before uploading index.json')
            sleep(1)
        logger.debug(f'Queue size: {self.executor._work_queue.qsize()}. Uploading ' +
                     f'index.json to {self.remote}')
        future = self.executor.submit(self.cloud_writer.upload_file, basename)
        future.add_done_callback(self.exception_callback)

    def finish(self) -> None:
        """Finish writing samples."""
        if self.event.is_set():
            # If an error occurred, cancel any outstanding uploads
            self.cancel_future_jobs()
        else:
            # Otherwise finish any remaining uploads and write index, and wait for completion
            if self.new_samples:
                self.flush_shard()
                self._reset_cache()
            self._write_index()
            logger.debug(f'Waiting for all shard files to get uploaded to {self.remote}')
            self.executor.shutdown(wait=True)

        if self.remote and not self.keep_local:
            shutil.rmtree(self.local, ignore_errors=True)

        # Final check, in case error occurred in an upload during shutdown
        if self.event.is_set():
            raise Exception('One of the threads failed. Check other traceback for more ' +
                            'details.')

    def cancel_future_jobs(self) -> None:
        """Shutting down the executor and cancel all the pending jobs."""
        # Beginning python v3.9, ThreadPoolExecutor.shutdown() has a new parameter `cancel_futures`
        self.executor.shutdown(wait=False, cancel_futures=True)

    def exception_callback(self, future: Future) -> None:
        """Raise an exception to the caller if exception generated by one of an async thread.

        Also, set the thread event to let other threads knows about the exception in one of the
        thread.

        Args:
            future (Future): Contains the status of the task

        Raises:
            exception: re-raise an exception
        """
        exception = future.exception()
        if exception:
            # Set the event to let other pool thread know about the exception
            self.event.set()
            # log exception; raising does not propagate to caller as this is called
            # from Future thread and would just be logged as an unexpected error
            logger.error(f'Exception in writer thread: {exception}')

    def __enter__(self) -> Self:
        """Enter context manager.

        Returns:
            Self: This object.
        """
        return self

    def __exit__(self, exc_type: Optional[type[BaseException]], exc: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        """Exit context manager.

        Args:
            exc_type (Type[BaseException], optional): Exc type.
            exc (BaseException, optional): Exc.
            traceback (TracebackType, optional): Traceback.
        """
        self.finish()


class JointWriter(Writer):
    """Writes a streaming dataset with joint shards.

    Args:
        out (str | Tuple[str, str]): Output dataset directory to save shard files.

            1. If ``out`` is a local directory, shard files are saved locally.
            2. If ``out`` is a remote directory, a local temporary directory is created to
               cache the shard files and then the shard files are uploaded to a remote
               location. At the end, the temp directory is deleted once shards are uploaded.
            3. If ``out`` is a tuple of ``(local_dir, remote_dir)``, shard files are saved in the
               `local_dir` and also uploaded to a remote location.
        keep_local (bool): If the dataset is uploaded, whether to keep the local dataset directory
            or remove it after uploading. Defaults to ``False``.
        compression (str, optional): Optional compression or compression:level. Defaults to
            ``None``.
        hashes (List[str], optional): Optional list of hash algorithms to apply to shard files.
            Defaults to ``None``.
        size_limit (int, optional): Optional shard size limit, after which point to start a new
            shard. If None, puts everything in one shard. Defaults to ``1 << 26``.
        extra_bytes_per_shard (int): Extra bytes per serialized shard (for computing shard size
            while writing). Defaults to ``0``.
        extra_bytes_per_sample (int): Extra bytes per serialized sample (for computing shard size
            while writing). Defaults to ``0``.
        **kwargs (Any): Additional settings for the Writer.

            progress_bar (bool): Display TQDM progress bars for uploading output dataset files to
                a remote location. Default to ``False``.
            max_workers (int): Maximum number of threads used to upload output dataset files in
                parallel to a remote location. One thread is responsible for uploading one shard
                file to a remote location. Default to ``min(32, (os.cpu_count() or 1) + 4)``.
            retry (int): Number of times to retry uploading a file to a remote location.
                Default to ``2``.
    """

    def __init__(self,
                 *,
                 out: Union[str, tuple[str, str]],
                 keep_local: bool = False,
                 compression: Optional[str] = None,
                 hashes: Optional[list[str]] = None,
                 size_limit: Optional[Union[int, str]] = 1 << 26,
                 extra_bytes_per_shard: int = 0,
                 extra_bytes_per_sample: int = 0,
                 **kwargs: Any) -> None:
        super().__init__(out=out,
                         keep_local=keep_local,
                         compression=compression,
                         hashes=hashes,
                         size_limit=size_limit,
                         extra_bytes_per_shard=extra_bytes_per_shard,
                         extra_bytes_per_sample=extra_bytes_per_sample,
                         **kwargs)

    @abstractmethod
    def encode_joint_shard(self) -> bytes:
        """Encode a joint shard out of the cached samples (single file).

        Returns:
            bytes: File data.
        """
        raise NotImplementedError

    def flush_shard(self) -> None:

        raw_data_basename, zip_data_basename = self._name_next_shard()
        raw_data = self.encode_joint_shard()
        raw_data_info, zip_data_info = self._process_file(raw_data, raw_data_basename,
                                                          zip_data_basename)
        obj = {
            'samples': len(self.new_samples),
            'raw_data': raw_data_info,
            'zip_data': zip_data_info
        }
        obj.update(self.get_config())
        self.shards.append(obj)

        # Execute the task if there is no exception in any of the async threads.
        future = self.executor.submit(self.cloud_writer.upload_file, zip_data_basename or
                                      raw_data_basename)
        future.add_done_callback(self.exception_callback)


class SplitWriter(Writer):
    """Writes a streaming dataset with split shards.

    Split shards refer to raw data (csv, json, etc.) paired with an index into it.

    Args:
        out (str | Tuple[str, str]): Output dataset directory to save shard files.

            1. If ``out`` is a local directory, shard files are saved locally.
            2. If ``out`` is a remote directory, a local temporary directory is created to
               cache the shard files and then the shard files are uploaded to a remote
               location. At the end, the temp directory is deleted once shards are uploaded.
            3. If ``out`` is a tuple of ``(local_dir, remote_dir)``, shard files are saved in the
               `local_dir` and also uploaded to a remote location.
        keep_local (bool): If the dataset is uploaded, whether to keep the local dataset directory
            or remove it after uploading. Defaults to ``False``.
        compression (str, optional): Optional compression or compression:level. Defaults to
            ``None``.
        hashes (List[str], optional): Optional list of hash algorithms to apply to shard files.
            Defaults to ``None``.
        size_limit (int, optional): Optional shard size limit, after which point to start a new
            shard. If None, puts everything in one shard. Defaults to ``1 << 26``.
        **kwargs (Any): Additional settings for the Writer.

            progress_bar (bool): Display TQDM progress bars for uploading output dataset files to
                a remote location. Default to ``False``.
            max_workers (int): Maximum number of threads used to upload output dataset files in
                parallel to a remote location. One thread is responsible for uploading one shard
                file to a remote location. Default to ``min(32, (os.cpu_count() or 1) + 4)``.
            retry (int): Number of times to retry uploading a file to a remote location.
                Default to ``2``.
    """

    extra_bytes_per_shard = 0
    extra_bytes_per_sample = 0

    def __init__(self,
                 *,
                 out: Union[str, tuple[str, str]],
                 keep_local: bool = False,
                 compression: Optional[str] = None,
                 hashes: Optional[list[str]] = None,
                 size_limit: Optional[Union[int, str]] = 1 << 26,
                 **kwargs: Any) -> None:
        super().__init__(out=out,
                         keep_local=keep_local,
                         compression=compression,
                         hashes=hashes,
                         size_limit=size_limit,
                         extra_bytes_per_shard=self.extra_bytes_per_shard,
                         extra_bytes_per_sample=self.extra_bytes_per_sample,
                         **kwargs)

    @abstractmethod
    def encode_split_shard(self) -> tuple[bytes, bytes]:
        """Encode a split shard out of the cached samples (data, meta files).

        Returns:
            Tuple[bytes, bytes]: Data file, meta file.
        """
        raise NotImplementedError

    def flush_shard(self) -> None:

        raw_data_basename, zip_data_basename = self._name_next_shard()
        raw_meta_basename, zip_meta_basename = self._name_next_shard('meta')
        raw_data, raw_meta = self.encode_split_shard()
        raw_data_info, zip_data_info = self._process_file(raw_data, raw_data_basename,
                                                          zip_data_basename)
        raw_meta_info, zip_meta_info = self._process_file(raw_meta, raw_meta_basename,
                                                          zip_meta_basename)
        obj = {
            'samples': len(self.new_samples),
            'raw_data': raw_data_info,
            'zip_data': zip_data_info,
            'raw_meta': raw_meta_info,
            'zip_meta': zip_meta_info
        }
        obj.update(self.get_config())
        self.shards.append(obj)

        # Execute the task if there is no exception in any of the async threads.
        future = self.executor.submit(self.cloud_writer.upload_file, zip_data_basename or
                                      raw_data_basename)
        future.add_done_callback(self.exception_callback)

        # Execute the task if there is no exception in any of the async threads.
        future = self.executor.submit(self.cloud_writer.upload_file, zip_meta_basename or
                                      raw_meta_basename)
        future.add_done_callback(self.exception_callback)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Apportion shards/samples to nodes/ranks/workers for elastically deterministic sample order."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

from streaming.base.batching.device_per_stream import generate_work_device_per_stream_batching
from streaming.base.batching.per_stream import generate_work_per_stream_batching
from streaming.base.batching.random import generate_work_random_batching
from streaming.base.batching.stratified import generate_work_stratified_batching
from streaming.base.world import World

if TYPE_CHECKING:
    from streaming.base.dataset import StreamingDataset

batching_methods = {
    'random': generate_work_random_batching,
    'stratified': generate_work_stratified_batching,
    'per_stream': generate_work_per_stream_batching,
    'device_per_stream': generate_work_device_per_stream_batching,
}


def generate_work(batching_method: str, dataset: StreamingDataset, world: World, epoch: int,
                  sample_in_epoch: int) -> NDArray[np.int64]:
    """Apportion shards/samples to nodes/ranks/workers for elastically deterministic sample order.

    Args:
        batching_method (str): The batching method to use.
        dataset (StreamingDataset): Dataset to generate the partition for.
        world (World): World state.
        epoch (int): Which epoch it is.
        sample_in_epoch (int): Where we are in the epoch.

    Returns:
        NDArray[np.int64]: The epoch (num physical nodes, ranks per node, workers per rank,
            batches per worker, batch size).
    """
    get = batching_methods[batching_method]
    return get(dataset, world, epoch, sample_in_epoch)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

from filecmp import dircmp
from typing import Any, Optional, Union

import numpy as np
import pytest

from streaming.base import StreamingDataset
from streaming.base.compression import (Brotli, Bzip2, Gzip, Snappy, Zstandard, compress,
                                        decompress, get_compression_extension, is_compression)
from tests.common.datasets import SequenceDataset, write_mds_dataset


class TestBrotli:

    def test_constructor(self):
        brotli = Brotli()
        assert brotli.level == 11

    def test_extension(self):
        brotli = Brotli()
        assert brotli.extension == 'br'

    def test_levels(self):
        brotli = Brotli()
        levels = list(range(12))
        assert brotli.levels == levels

    @pytest.mark.parametrize('level', list(range(12)))
    @pytest.mark.parametrize(
        'data',
        [
            str.encode('Hello World 789*!' * 10, encoding='utf-8'),
            np.int64(1234).tobytes(),
            np.random.randint(0, 255, (32, 32, 3)).tobytes(),
        ],
        ids=['string', 'int', 'image'],
    )
    def test_comp_decomp(self, level: int, data: bytes):
        brotli = Brotli(level)
        output = brotli.decompress(brotli.compress(data))
        assert output == data

    @pytest.mark.parametrize('data', [100, 1.2, 'bigdata'])
    def test_invalid_data(self, data: Any):
        brotli = Brotli()
        with pytest.raises(TypeError):
            _ = brotli.compress(data)


class TestBzip2:

    def test_constructor(self):
        bzip2 = Bzip2()
        assert bzip2.level == 9

    def test_extension(self):
        bzip2 = Bzip2()
        assert bzip2.extension == 'bz2'

    def test_levels(self):
        bzip2 = Bzip2()
        levels = list(range(1, 10))
        assert bzip2.levels == levels

    @pytest.mark.parametrize('level', list(range(1, 10)))
    @pytest.mark.parametrize(
        'data',
        [
            str.encode('Hello World 789*!' * 10, encoding='utf-8'),
            np.int64(1234).tobytes(),
            np.random.randint(0, 255, (32, 32, 3)).tobytes(),
        ],
        ids=['string', 'int', 'image'],
    )
    def test_comp_decomp(self, level: int, data: bytes):
        bzip2 = Bzip2(level)
        output = bzip2.decompress(bzip2.compress(data))
        assert output == data

    @pytest.mark.parametrize('data', [100, 1.2, 'bigdata'])
    def test_invalid_data(self, data: Any):
        bzip2 = Bzip2()
        with pytest.raises(TypeError):
            _ = bzip2.compress(data)


class TestGzip:

    def test_constructor(self):
        gzip = Gzip()
        assert gzip.level == 9

    def test_extension(self):
        gzip = Gzip()
        assert gzip.extension == 'gz'

    def test_levels(self):
        gzip = Gzip()
        levels = list(range(10))
        assert gzip.levels == levels

    @pytest.mark.parametrize('level', list(range(10)))
    @pytest.mark.parametrize(
        'data',
        [
            str.encode('Hello World 789*!' * 10, encoding='utf-8'),
            np.int64(1234).tobytes(),
            np.random.randint(0, 255, (32, 32, 3)).tobytes(),
        ],
        ids=['string', 'int', 'image'],
    )
    def test_comp_decomp(self, level: int, data: bytes):
        gzip = Gzip(level)
        output = gzip.decompress(gzip.compress(data))
        assert output == data

    @pytest.mark.parametrize('data', [100, 1.2, 'bigdata'])
    def test_invalid_data(self, data: Any):
        gzip = Gzip()
        with pytest.raises(TypeError):
            _ = gzip.compress(data)


class TestSnappy:

    def test_extension(self):
        snappy = Snappy()
        assert snappy.extension == 'snappy'

    @pytest.mark.parametrize(
        'data',
        [
            str.encode('Hello World 789*!' * 10, encoding='utf-8'),
            np.int64(1234).tobytes(),
            np.random.randint(0, 255, (32, 32, 3)).tobytes(),
        ],
        ids=['string', 'int', 'image'],
    )
    def test_comp_decomp(self, data: bytes):
        snappy = Snappy()
        output = snappy.decompress(snappy.compress(data))
        assert output == data

    @pytest.mark.parametrize('data', [100, 1.2])
    def test_invalid_data(self, data: Any):
        snappy = Snappy()
        with pytest.raises(TypeError):
            _ = snappy.compress(data)


class TestZstandard:

    def test_constructor(self):
        zstd = Zstandard()
        assert zstd.level == 3

    def test_extension(self):
        zstd = Zstandard()
        assert zstd.extension == 'zstd'

    def test_levels(self):
        zstd = Zstandard()
        levels = list(range(1, 23))
        assert zstd.levels == levels

    @pytest.mark.parametrize('level', list(range(1, 23)))
    @pytest.mark.parametrize(
        'data',
        [
            str.encode('Hello World 789*!' * 10, encoding='utf-8'),
            np.int64(1234).tobytes(),
            np.random.randint(0, 255, (32, 32, 3)).tobytes(),
        ],
        ids=['string', 'int', 'image'],
    )
    def test_comp_decomp(self, level: int, data: bytes):
        zstd = Zstandard(level)
        output = zstd.decompress(zstd.compress(data))
        assert output == data

    @pytest.mark.parametrize('data', [100, 1.2, 'bigdata'])
    def test_invalid_data(self, data: Any):
        zstd = Zstandard()
        with pytest.raises(TypeError):
            _ = zstd.compress(data)


@pytest.mark.parametrize(('algo', 'is_valid'), [('br', True), ('br:7', True), ('bz2', True),
                                                ('gz', True), ('gz:4', True), ('snappy', True),
                                                ('zstd', True), ('xyz', False)])
def test_is_compression(algo: str, is_valid: bool):
    is_algo_valid = is_compression(algo)
    assert is_algo_valid == is_valid


@pytest.mark.parametrize('algo', ['br', 'bz2', 'gz', 'snappy', 'zstd'])
def test_success_get_compression_extension(algo: str):
    extension = get_compression_extension(algo)
    assert extension == algo


@pytest.mark.parametrize('algo', ['xyz'])
def test_invalid_compression_extension(algo: str):
    with pytest.raises(ValueError) as exc_info:
        _ = get_compression_extension(algo)
    assert exc_info.match(r'.*is not a supported compression algorithm.*')


@pytest.mark.parametrize(
    ('algo', 'data', 'expected_data'), [('br:1', b'hello', b'\x0b\x02\x80hello\x03'),
                                        (None, b'hello', b'hello')])
def test_compress(algo: Optional[str], data: bytes, expected_data: bytes):
    output = compress(algo, data)
    assert output == expected_data


def test_compress_invalid_compression_algo():
    with pytest.raises(ValueError) as exc_info:
        _ = compress('br:99', b'hello')
    assert exc_info.match(r'.*is not a supported compression algorithm.*')


@pytest.mark.parametrize(('algo', 'data', 'expected_data'),
                         [('br:1', b'\x0b\x02\x80hello\x03', b'hello'),
                          (None, b'\x0b\x02\x80hello\x03', b'\x0b\x02\x80hello\x03')])
def test_decompress(algo: Optional[str], data: bytes, expected_data: bytes):
    output = decompress(algo, data)
    assert output == expected_data


def test_decompress_invalid_compression_algo():
    with pytest.raises(ValueError) as exc_info:
        _ = decompress('gz:99', b'\x0b\x02\x80hello\x03')
    assert exc_info.match(r'.*is not a supported compression algorithm.*')


def check_for_diff_files(dir: dircmp, compression_ext: Union[None, str]):
    """Check recursively for different files in a dircmp object.

    Local directory also has the uncompressed files, ignore it during file comparison.
    """
    if compression_ext:
        for file in dir.diff_files:
            assert not file.endswith(compression_ext)
    else:
        assert len(dir.diff_files) == 0
    for subdir in dir.subdirs:
        check_for_diff_files(subdir, compression_ext)


@pytest.mark.parametrize('compression', [None, 'br:11', 'bz2:9', 'gz:5', 'snappy', 'zstd:15'])
@pytest.mark.parametrize('num_samples', [9867])
@pytest.mark.parametrize('size_limit', [16_384])
def test_dataset_compression(compressed_local_remote_dir: tuple[str, str, str],
                             compression: Optional[str], num_samples: int, size_limit: int):
    shuffle = True
    compressed, local, remote = compressed_local_remote_dir
    samples = SequenceDataset(num_samples)
    columns = dict(zip(samples.column_names, samples.column_encodings))
    compression_ext = compression.split(':')[0] if compression else None

    write_mds_dataset(out_root=compressed,
                      columns=columns,
                      samples=samples,
                      size_limit=size_limit,
                      compression=compression)
    samples = SequenceDataset(num_samples)
    write_mds_dataset(out_root=remote,
                      columns=columns,
                      samples=samples,
                      size_limit=size_limit,
                      compression=None)

    dataset = StreamingDataset(local=local, remote=compressed, shuffle=shuffle, batch_size=1)

    for _ in dataset:
        pass  # download sample

    dcmp = dircmp(remote, local)
    check_for_diff_files(dcmp, compression_ext)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Script for simulating training downloads and throughput, and displaying results."""

import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import humanize
from core.create_index import create_stream_index
from core.main import simulate
from core.sim_dataset import SimulationDataset
from core.sim_time import TimeUnit, ensure_time
from core.utils import get_simulation_stats
from interfaces.interface_utils import plot_simulation

from streaming.base import Stream

# Input Parameters

# dataset
shards = 20850  # number of shards
samples_per_shard = 4093  # number of samples per shard
avg_raw_shard_size = 67092639  # average shard size (bytes)
avg_zip_shard_size = 15000000  # average compressed shard size (bytes)

# training
max_duration = '1000ba'  # max duration of training (batches: "ba", epochs: "ep")
epoch_size = None  # epoch size (samples)
device_batch_size = 16  # device batch size (samples)

# streaming
workers = 8  # number of workers per device
canonical_nodes = 2  # number of canonical nodes
predownload = 32  # number of samples to predownload per worker (samples)
cache_limit = None  # cache limit (bytes)
shuffle = True  # whether to shuffle dataset
shuffle_algo = 'py1b'  # shuffling algorithm
shuffle_block_size = 16000000  # shuffling block size (samples)
seed = 17  # random seed

# hardware and network
physical_nodes = 2  # number of physical nodes
devices = 8  # number of devices per node
time_per_sample = 0.0175  # time to process one sample on one device (seconds)
node_internet_bandwidth = 1e7  # network internet per node (bytes/s)

# ---------------------------------------------- #

# instantiate SimulationDataset on the same parameters for the new simulation function

stream_indexpath = create_stream_index(shards, samples_per_shard, avg_raw_shard_size,
                                       avg_zip_shard_size)
stream_folder = os.path.dirname(stream_indexpath)
stream = Stream(local=stream_folder)
max_duration = ensure_time(max_duration, TimeUnit.EPOCH)

dataset = SimulationDataset(nodes=physical_nodes,
                            devices=devices,
                            workers=workers,
                            streams=[stream],
                            epoch_size=epoch_size,
                            predownload=predownload,
                            cache_limit=cache_limit,
                            num_canonical_nodes=canonical_nodes,
                            batch_size=device_batch_size,
                            shuffle=True,
                            shuffle_algo=shuffle_algo,
                            shuffle_seed=seed,
                            shuffle_block_size=shuffle_block_size)

node_internet_bandwidth = int(node_internet_bandwidth)
results = next(
    simulate(dataset=dataset,
             time_per_sample=time_per_sample,
             node_network_bandwidth=node_internet_bandwidth,
             max_duration=max_duration))

if len(results) != 4:
    raise ValueError(f'Simulation with generate=False should return 4 final results. ' +
                     f'Instead, received `results` of length {len(results)}.')
step_times, step_downloads, startup_time, min_cache_limit = results
global_batch_size = device_batch_size * devices * physical_nodes

# Display simulation stats
total_batches = len(step_times)
all_throughput_drops, warmup_time, warmup_step, post_warmup_throughput_drops = \
    get_simulation_stats(step_times, time_per_sample, global_batch_size//(physical_nodes*devices))
print('\nSimulation Stats:')
print(f'Minimum cache limit needed: {humanize.naturalsize(min_cache_limit)}')
if cache_limit is not None and cache_limit < min_cache_limit:
    # Cache limit is too low, and will cause shard redownloads / throughput drops.
    print('⚠️ The provided cache limit is lower than the minimum cache limit needed to \
          prevent shard re-downloads. This can cause throughput issues.')
if warmup_step == total_batches:
    # display error if the warmup phase is the whole run, so we never hit peak throughput.
    print('🚨 This configuration is severely bottlenecked by downloading. The run will not be \
            performant.')
elif post_warmup_throughput_drops:
    # display warning if post-warmup throughput drops are more than 10% of the run.
    print(
        '⚠️ This configuration experiences some downloading-related slowdowns even after warmup.')
print('{0} steps, or {1:.1f}% of all steps, waited for shard downloads.'\
      .format(all_throughput_drops, 100 * all_throughput_drops / (total_batches)))
if warmup_step != total_batches:
    # only display post-warmup throughput drop info if we actually ended the warmup period (i.e. we hit peak throughput at some point)
    print('There were {} steps that waited for shard downloads after the warmup period.'\
          .format(post_warmup_throughput_drops))
print('Estimated time to first batch: {0:.2f} s'.format(startup_time))
print('Estimated warmup time: {0:.2f} s'.format(warmup_time))

plot_simulation(step_times, step_downloads)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Shuffling algorithm that shuffles intra-shard in one place.

This algorithm is roughly twice as fast as algorithm ``py2s``, and ever so slightly biased.

Bias in this case merely refers to how we assign samples when we split shards at canonical node
boundaries, which is non-random in this algorithm. In practice, we found this does not matter to
convergence, while making us faster.
"""

import numpy as np
from numpy.typing import NDArray


def divide_spans(spans: list[tuple[int, int]], num_samples: int, num_parts: int) -> \
        tuple[list[tuple[int, int]], list[tuple[int, int]]]:
    """Divide the spans into discrete, equal sized partitions.

    Don't use ``spans`` after this, as it is modified in-place for performance reasons.

    Args:
        spans (List[Tuple[int, int]]): List of spans to partition.
        num_samples (int): Total number of samples across all spans.
        num_parts (int): Number of groupings to divide spans into.

    Returns:
        Tuple[List[Tuple, int, int]], List[Tuple[int, int]]]: Spans and super spans.
    """
    begin_part = 0
    span_index = 0
    samples_so_far = 0

    out_spans = []
    super_spans = []

    for part in range(num_parts):
        # note that the size of a part (canonical node) is num_samples // num_parts.
        part_end = num_samples * (part + 1) // num_parts

        # loop over spans until we've filled up our part (canonical node) completely
        while True:
            if span_index == len(spans):
                break

            # input spans are the shard spans. these can be unequally sized and may cross
            # part (canonical node) boundaries.
            span = spans[span_index]
            # spans are (begin, end excl)
            samples_this_span = span[1] - span[0]
            # check if the shard span contains more samples than the part (canonical node) can fit
            if part_end < samples_so_far + samples_this_span:
                # if there is space left in the part, split the span
                if samples_so_far < part_end:
                    split = part_end - samples_so_far
                    # create a span, filling up with as many samples as possible from shard span
                    new_span = span[0], span[0] + split
                    out_spans.append(new_span)
                    # modify the old shard span to reflect that it's been split
                    spans[span_index] = span[0] + split, span[1]
                    samples_so_far += split
                break

            out_spans.append(span)
            span_index += 1
            samples_so_far += samples_this_span

        # super spans are tell us which new spans belong to each part (canonical node)
        # as a tuple of (begin span index, end span index excl)
        super_span = begin_part, len(out_spans)
        super_spans.append(super_span)
        begin_part = len(out_spans)

    return out_spans, super_spans


def get_shuffle_py1s(shard_sizes: NDArray[np.int64],
                     num_canonical_nodes: int,
                     seed: int,
                     epoch: int,
                     block_size: int = 1 << 18) -> NDArray[np.int64]:
    """Get the shuffled global ordering of samples for an epoch.

    The assignment of shards to nodes is fixed across epochs, but each grouping of shards is
    processed concurrently in a different order by each node's workers each epoch.

    Args:
        shard_sizes (NDArray[np.int64]): Number of samples contained in each shard, in order.
        num_canonical_nodes (int): Number of canonical nodes.
        seed (int): Base random seed, which is held constant over an entire training run.
        epoch (int): Current epoch, which is added to the seed to get a different deterministic
            shuffle each epoch.
        block_size (int): Unit of shuffle (ignored, because we shuffle on the basis of shards).
            Defaults to ``1 << 18``.

    Returns:
        NDArray[np.int64]: 1:1 mapping of sample ID to shuffled sample ID.
    """
    # Create each shard's sample ID span (begin, end excl).
    spans = []
    num_samples = 0
    for shard_size in shard_sizes:
        span = num_samples, num_samples + shard_size
        spans.append(span)
        num_samples += shard_size

    # Generate the initial ordering of shards, which is fixed over an entire training run.
    run_rng = np.random.default_rng(seed)
    run_rng.shuffle(spans)

    # Break the shard spans at canonical node boundaries.
    spans, super_spans = divide_spans(spans, num_samples, num_canonical_nodes)

    # Shuffle the span ordering within each canonical node uniquely to this epoch.
    epoch_rng = np.random.default_rng(seed + epoch)
    for begin, end in super_spans:
        part = spans[begin:end]
        epoch_rng.shuffle(part)  # pyright: ignore
        spans[begin:end] = part

    # Populate the global sample ID mapping, shuffling within each span.
    ids = np.empty(num_samples, np.int64)
    offset = 0
    for begin, end in spans:
        span_size = end - begin
        ids[offset:offset + span_size] = np.arange(begin, end)
        epoch_rng.shuffle(ids[offset:offset + span_size])
        offset += span_size

    return ids



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Natively supported CV datasets."""

from streaming.vision.ade20k import StreamingADE20K as StreamingADE20K
from streaming.vision.cifar10 import StreamingCIFAR10 as StreamingCIFAR10
from streaming.vision.coco import StreamingCOCO as StreamingCOCO
from streaming.vision.imagenet import StreamingImageNet as StreamingImageNet

__all__ = ['StreamingADE20K', 'StreamingCIFAR10', 'StreamingCOCO', 'StreamingImageNet']



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""A numpy array of predetermined shape and dtype that lives in shared memory."""

from typing import Any, Union

import numpy as np
from numpy.typing import NDArray

from streaming.base.shared.memory import SharedMemory


class SharedArray:
    """A numpy array of predetermined shape and dtype that lives in shared memory.

    Args:
        shape (Union[int, Tuple[int]]): Shape of the array.
        dtype (type): Dtype of the array.
        name (str): Its name in shared memory.
    """

    def __init__(self, shape: Union[int, tuple[int]], dtype: type, name: str) -> None:
        self.shape = np.empty(shape).shape
        self.dtype = dtype
        self.name = name
        size = int(np.prod(shape) * dtype(0).nbytes)
        self.shm = SharedMemory(name=name, size=size)

    def numpy(self) -> NDArray:
        """Get as a numpy array.

        We can't just internally store and use this numpy array shared memory wrapper because it's
        not compatible with spawn.
        """
        return np.ndarray(self.shape, buffer=self.shm.buf, dtype=self.dtype)

    def __len__(self) -> int:
        """Get the length (i.e., size along the first axis).

        Returns:
            int: The length.
        """
        return int(self.shape[0])

    def __getitem__(self, index: Any) -> Any:
        """Get the scalar(s) at the given index, slice, or array of indices.

        Args:
            index (Any): The index, slice, or array of indices.

        Returns:
            The scalar(s) at the given location(s).
        """
        arr = self.numpy()
        return arr[index]

    def __setitem__(self, index: Any, value: Any) -> Any:
        """Set the scalar(s) at the given index, slice, or array of indices.

        Args:
            index (Any): The index, slice, or array of indices.
            value (Any): The scalar(s) at the given location(s).
        """
        arr = self.numpy()
        arr[index] = value



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Shuffling algorithm that naively shuffles all-to-all.

Useful for single-node training on small data, where you want the most random shuffle possible.

Statistically, this algorithm will result in all nodes downloading all shards, with those downloads
all happening at the start of the epoch, bringing training to a crawl.
"""

import numpy as np
from numpy.typing import NDArray


def get_shuffle_naive(shard_sizes: NDArray[np.int64],
                      num_canonical_nodes: int,
                      seed: int,
                      epoch: int,
                      block_size: int = 1 << 18) -> NDArray[np.int64]:
    """Get the shuffled global ordering of samples for an epoch.

    The assignment of shards to nodes is fixed across epochs, but each grouping
    of shards is processed concurrently in a different order by each node's
    workers each epoch.

    Args:
        shard_sizes (NDArray[np.int64]): Number of samples contained in each shard, in order.
        num_canonical_nodes (int): Number of canonical nodes.
        seed (int): Base random seed, which is held constant over an entire training run.
        epoch (int): Current epoch, which is added to the seed to get a different deterministic
            shuffle each epoch.
        block_size (int): Unit of shuffle (ignored because we shuffle all samples together).
            Defaults to ``1 << 18``.

    Returns:
        NDArray[np.int64]: 1:1 mapping of sample ID to shuffled sample ID.
    """
    rng = np.random.default_rng(seed + epoch)
    return rng.permutation(sum(shard_sizes))



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Convert an MP4-inside MDS dataset to an MP4-outside MDS dataset."""

import os
from argparse import ArgumentParser, Namespace

from tqdm import tqdm

from streaming import MDSWriter, StreamingDataset

out_columns = {
    'videoid': 'str',
    'name': 'str',
    'page_idx': 'str',
    'page_dir': 'str',
    'duration': 'int',
    'contentUrl': 'str',
    'content_path': 'str',
}


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Returns:
        Namespace: Command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument(
        '--in',
        type=str,
        required=True,
        help='Input mp4-inside MDS dataset directory',
    )
    args.add_argument(
        '--out_mds',
        type=str,
        required=True,
        help='Output mp4-outside MDS dataset directory',
    )
    args.add_argument(
        '--out_mp4',
        type=str,
        required=True,
        help='Output mp4 videos directory',
    )
    args.add_argument(
        '--compression',
        type=str,
        default='zstd:16',
        help='Compression',
    )
    args.add_argument(
        '--hashes',
        type=str,
        default='sha1,xxh3_64',
        help='Hashes',
    )
    args.add_argument(
        '--size_limit',
        type=int,
        default=1 << 25,
        help='Shard size limit',
    )
    return args.parse_args()


def main(args: Namespace) -> None:
    """Convert an MP4-inside MDS dataset to an MP4-outside MDS dataset.

    Args:
        args (Namespace): Command-line arguments.
    """
    hashes = args.hashes.split(',') if args.hashes else []
    dataset = StreamingDataset(local=getattr(args, 'in'), batch_size=1)
    with MDSWriter(out=args.out_mds,
                   columns=out_columns,
                   compression=args.compression,
                   hashes=hashes,
                   size_limit=args.size_limit) as out:
        for i, sample in tqdm(enumerate(dataset), total=dataset.num_samples):
            content_path = f'{i // 1000:03}/{i % 1000:03}.mp4'
            filename = os.path.join(args.out_mp4, content_path)
            dirname = os.path.dirname(filename)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(filename, 'wb') as out_mp4:
                out_mp4.write(sample['content'])
            del sample['content']
            sample['content_path'] = content_path
            out.write(sample)


if __name__ == '__main__':
    main(parse_args())
